# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .atoms import Helium, Nitrogen, Oxygen, Argon, Carbon
from .material import Material
from .database import db, _manually_add_material_to_db


# Liquid gases
LiquidHelium   = Helium.adapt(state='liquid',   density=0.146,  temperature=1.9,   pressure=None)
LiquidNitrogen = Nitrogen.adapt(state='liquid', density=0.808,  temperature=77.36, pressure=None)
LiquidOxygen   = Oxygen.adapt(state='liquid',   density=1.141,  temperature=90.20, pressure=None)
LiquidArgon    = Argon.adapt(state='liquid',    density=1.3954, temperature=87.3,  pressure=None)


# Carbon variants
Carbon140 = Carbon.adapt(density=1.40)
Carbon175 = Carbon.adapt(density=1.75)
Carbon180 = Carbon.adapt(density=1.80)
Carbon185 = Carbon.adapt(density=1.85)
Diamond   = Carbon.adapt(density=3.52, excitation_energy=88.5)
GraphiteR4550 = Carbon.adapt(density=1.83, excitation_energy=78.0)
CarbonFibreCarbon = GraphiteR4550.adapt(density=1.67, nuclear_radius=0.25, nuclear_elastic_slope=70.0,
                                        cross_section=[0.337, 0.232, 0, 0, 0, 0.0076e-2], hcut=0.02,
                                        crystal_plane_distance=0.63e-7, eta=0.9,
                                        crystal_potential=21.0, nuclear_collision_length=1.e-12)

# Metadata for database
# =====================

_manually_add_material_to_db(LiquidHelium,      'LiquidHelium')
_manually_add_material_to_db(LiquidNitrogen,    'LiquidNitrogen')
_manually_add_material_to_db(LiquidOxygen,      'LiquidOxygen')
_manually_add_material_to_db(LiquidArgon,       'LiquidArgon')
_manually_add_material_to_db(Carbon140,         'Carbon140',                           fluka_name='CC_1_40')
_manually_add_material_to_db(Carbon175,         'Carbon175',                           fluka_name='CC_1_75')
_manually_add_material_to_db(Carbon180,         'Carbon180',                           fluka_name='CC_1_80')
_manually_add_material_to_db(Carbon185,         'Carbon185',                           fluka_name='CC_1_85')
_manually_add_material_to_db(Diamond,           'Diamond')
_manually_add_material_to_db(GraphiteR4550,     'GraphiteR4550',                       fluka_name='GRAR4550')
_manually_add_material_to_db(CarbonFibreCarbon, 'CarbonFibreCarbon', short_name='CFC', fluka_name='AC150GPH')


Carbon140.info         = 'Carbon-based material. Used in LHC dump.'
Carbon175.info         = 'Carbon-based material. Used in LHC dump.'
Carbon180.info         = 'Carbon-based material. Used in LHC dump.'
Carbon185.info         = 'Carbon-based material. Used in LHC dump.'
GraphiteR4550.info     = 'Graphite material. Was a candidate for LHC collimators.'
CarbonFibreCarbon.info = 'Carbon-fibre composite. Used in LHC collimators.'


# Clean up namespace
del Helium, Nitrogen, Oxygen, Argon, Carbon
del Material
del db



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .material import Material
from ..pretty_print import style, pad_styled


class MaterialsDatabase:
    """Class to handle the materials database for Xcoll."""

    def __init__(self):
        self._materials = {}    # Case-insensitive keys. Each entry is a dict with 'material' and 'name' keys; the latter is meant to keep formatting for easier recognition when printing
        self._aliases = {}      # Case-insensitive keys, aliases to _materials. Each entry is a dict with 'refname' and 'name' keys; the latter is meant to keep formatting for easier recognition when printing
        self._fluka_names = {}  # Case-sensitive keys, aliases to _materials
        self._geant4_names = {} # Case-sensitive keys, aliases to _materials
        self._iter_index = 0
        self._iter_names = []

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))} (use .show() " \
             + f"to see the contents)>"

    def __str__(self):
        return self._str(format=False)

    def show(self, full=False):
        """Print the content of the accessor."""
        print(self._str(format=True, full=full))

    def _str(self, format, full=True):
        res = []
        padding = 36
        for typ in ["Atomic Elements", "Compounds", "Mixtures",
                    "Old Sixtrack Materials"]:
            header  = style(f'{typ}', bold=True, colour='navy', enabled=format)
            if typ != "Old Sixtrack Materials":
            #     header += style('   (still default from colldb)', dim=True, colour='navy', enabled=format)
            # else:
                header += style('  (aliases', italic=True, colour='navy', enabled=format)
                if full:
                    header += style(' <case-insensitive>', dim=True, colour='navy', enabled=format)
                header += style(' | ', italic=True, colour='navy', enabled=format)
                header += style('FLUKA-only aliases', italic=True, colour='forest_green', enabled=format)
                if full:
                    header += style(' <case-sensitive>', dim=True, colour='navy', enabled=format)
                header += style(' | ', italic=True, colour='navy', enabled=format)
                header += style('Geant4-only aliases', italic=True, colour='crimson', enabled=format)
                if full:
                    header += style(' <case-sensitive>', dim=True, colour='navy', enabled=format)
                header += style(')', italic=True, colour='navy', enabled=format)
            res.append(header)
            for name, mat in self._materials.items():
                this_mat = mat['material']
                if typ == "Atomic Elements" and (not this_mat.is_elemental
                or this_mat.name.startswith('K2')):
                        continue
                if typ == "Compounds" and (not this_mat.is_compound
                or this_mat.name.startswith('K2')):
                        continue
                if typ == "Mixtures" and (not this_mat.is_mixture
                or this_mat.name.startswith('K2')):
                        continue
                if typ == "Old Sixtrack Materials" \
                and not this_mat.name.startswith('K2'):
                    continue
                mess_name = style(mat['name'], bold=True, colour='navy', enabled=format)
                names = []
                names_stripped = [self._strip(mat['name'])]
                fluka_names = []
                geant4_names = []
                for refname in self._aliases.values():
                    if refname['refname'] == name:
                        names.append(style(refname['name'], italic=True, colour='navy', enabled=format))
                        names_stripped.append(self._strip(refname['name']))
                for fluka_name, refname in self._fluka_names.items():
                    if refname == name:
                        if self._strip(fluka_name) not in names_stripped:
                            fluka_names.append(style(fluka_name, italic=True, colour='forest_green', enabled=format))
                for geant4_name, refname in self._geant4_names.items():
                    if refname == name:
                        if self._strip(geant4_name) not in names_stripped:
                            geant4_names.append(style(geant4_name, italic=True, colour='crimson', enabled=format))
                if names or fluka_names or geant4_names:
                    mess_name = pad_styled(mess_name, 16)
                    mess_name += style(' (', italic=True, colour='navy', enabled=format)
                    mess_name += style(', ', italic=True, colour='navy', enabled=format).join(names)
                    mess_name += style('|', italic=True, colour='navy', enabled=format)
                    mess_name += style(',', italic=True, colour='forest_green', enabled=format).join(fluka_names)
                    mess_name += style('|', italic=True, colour='navy', enabled=format)
                    mess_name += style(', ', italic=True, colour='crimson', enabled=format).join(geant4_names)
                    mess_name += style(')', italic=True, colour='navy', enabled=format)
                if full:
                    mess_name += ':'
                    res.append(f"  {pad_styled(mess_name, padding)}  {this_mat}")
                else:
                    res.append(f"  {mess_name}")
            res.append('')
        return '\n'.join(res[:-1])  # Remove last empty line

    @property
    def fluka(self):
        return MaterialsSubDatabase(self._fluka_names, self, 'FLUKA')

    @property
    def geant4(self):
        return MaterialsSubDatabase(self._geant4_names, self, 'Geant4')

    def __getitem__(self, name):
        # Check if in main database
        if self._strip(name) in self._materials:
            return self._get(name)
        # Check if in aliases
        elif self._strip(name) in self._aliases:
            aliased_name = self._get_alias(name)
            if aliased_name not in self._materials:
                raise KeyError(f"Broken database: Material with alias '{name}' "
                               f"not found in the database.")
            return self._get(aliased_name)
        # Check in FLUKA and/or Geant4 names
        elif name in self.fluka and name in self.geant4:
            if self.fluka[name] == self.geant4[name]:
                return self.fluka[name]
            raise KeyError(f"Material with name '{name}' is ambiguous: exists "
                           f"both as FLUKA and Geant4 name but points to different "
                           f"materials. Please use db.fluka['{name}'] or "
                           f"db.geant4['{name}']")
        elif name in self.fluka:
            return self.fluka[name]
        elif name in self.geant4:
            return self.geant4[name]
        # If not found anywhere, raise error
        raise KeyError(f"Material with name '{name}' not found in the database.")

    def __setitem__(self, name, material):
        material = self._resolve(material)
        # Check if name already in main database
        if self._strip(name) in self._materials:
            if self._get(name) == material:
                return  # Already in database
            raise ValueError(f"Material with name '{name}' already exists in the database.")
        # Check if name already in aliases
        elif self._strip(name) in self._aliases:
            if self._get(self._get_alias(name)) == material:
                return  # Already in database
            raise ValueError(f"Material with name '{name}' already exists in the database.")
        # Check if name already in FLUKA names
        elif self._strip(name) in [self._strip(nn) for nn in self.fluka.keys()]:
            for nn in self.fluka.keys():
                if self._strip(nn) == self._strip(name):
                    if self.fluka[nn] == material:
                        return  # Already in database
            raise ValueError(f"Material with name '{name}' already exists in the FLUKA sub-database.")
        # Check if name already in Geant4 names
        elif self._strip(name) in [self._strip(nn) for nn in self.geant4.keys()]:
            for nn in self.geant4.keys():
                if self._strip(nn) == self._strip(name):
                    if self.geant4[nn] == material:
                        return  # Already in database
            raise ValueError(f"Material with name '{name}' already exists in the Geant4 sub-database.")
        # Check if material already exists (by identity, not by value).
        existing_name = self._find_name_by_material(material)
        if existing_name is not None:
            self._set_alias(name, existing_name)
            return
        # Otherwise, add new material
        self._set(name, material)
        if material.name is None:
            material.name = name

    def remove_material(self, material):
        material = self._resolve(material)
        # Delete from the main database
        to_delete = []
        for name, mat in self.items():
            if mat == material:
                to_delete.append(name)
        for name in to_delete:
            del self._materials[name]
        # Delete aliases
        to_delete = []
        for alias, name in self._aliases.items():
            if name['refname'] not in self._materials:
                to_delete.append(alias)
        for alias in to_delete:
            del self._aliases[alias]
        # Delete FLUKA aliases
        to_delete = []
        for fluka_name, refname in self.fluka.names.items():
            if refname not in self._materials:
                to_delete.append(fluka_name)
        for fluka_name in to_delete:
            del self._fluka_names[fluka_name]
        # Delete Geant4 aliases
        to_delete = []
        for geant4_name, refname in self.geant4.names.items():
            if refname not in self._materials:
                to_delete.append(geant4_name)
        for geant4_name in to_delete:
            del self._geant4_names[geant4_name]

    def update_material(self, old_material, new_material):
        old_material = self._resolve(old_material)
        old_name = self._find_name_by_material(old_material)
        if old_name is None:
            raise ValueError("Material not found in database.")
        self._set(old_name, new_material)

    def rename_material(self, material, new_name):
        material = self._resolve(material)
        if new_name in self._materials:
            if self[new_name] == material:
                return  # Already has this name or alias
            raise ValueError(f"Material with name '{new_name}' already exists.")
        old_name = self._find_name_by_material(material)
        del self._materials[old_name]
        self._set(new_name, material)
        # Rename aliases
        to_adapt = []
        for alias, name in self._aliases.items():
            if name['refname'] == old_name:
                to_adapt.append(alias)
        for alias in to_adapt:
            self._aliases[alias]['refname'] = new_name
        # Rename FLUKA aliases
        to_adapt = []
        for fluka_name, refname in self.fluka.items():
            if refname == old_name:
                to_adapt.append(fluka_name)
        for fluka_name in to_adapt:
            self.fluka.names[fluka_name] = new_name
        # Rename Geant4 aliases
        to_adapt = []
        for geant4_name, refname in self.geant4.items():
            if refname == old_name:
                to_adapt.append(geant4_name)
        for geant4_name in to_adapt:
            self.geant4.names[geant4_name] = new_name

    def remove_alias(self, name):
        if self._strip(name) in self._aliases:
            del self._aliases[self._strip(name)]

    def rename_alias(self, name, new_name):
        if self._strip(name) not in self._aliases:
            raise KeyError(f"Alias '{name}' not found.")
        # Check if new_name already exists
        if self._strip(new_name) in self._aliases \
        or self._strip(new_name) in self._materials:
            if self[new_name] == self[name]:
                del self._aliases[self._strip(name)]
                return  # Already has this name
            raise ValueError(f"Material with name '{new_name}' already exists.")
        refname = self._aliases.pop(self._strip(name))['refname']
        self._aliases[self._strip(new_name)] = {'name': new_name, 'refname': refname}

    def __contains__(self, name):
        if name is None:
            return False
        elif isinstance(name, Material):
            return name in self.values()
        elif isinstance(name, str) and self._strip(name) in self._materials:
            return True
        elif isinstance(name, str) and self._strip(name) in self._aliases:
            return True
        elif isinstance(name, str) and self._strip(name) in [self._strip(nn) for nn in self.fluka.keys()]:
            return True
        elif isinstance(name, str) and self._strip(name) in [self._strip(nn) for nn in self.geant4.keys()]:
            return True
        else:
            return False

    def __iter__(self):
        self._iter_index = 0
        self._iter_names = list(self._materials.keys())
        return self

    def __next__(self):
        if self._iter_index >= len(self._iter_names):
            raise StopIteration
        value = self._materials[self._iter_names[self._iter_index]]['material']
        self._iter_index += 1
        return value

    def __len__(self):
        return len(self._materials)

    def items(self):
        for name, mat in self._materials.items():
            yield name, mat['material']

    def keys(self):
        for name in self._materials.keys():
            yield name

    def values(self):
        for mat in self._materials.values():
            yield mat['material']

    # ----- Helpers -----

    def _strip(self, name):
        return name.strip().lower().replace(' ', '').replace('-', '').replace('_', '')

    def _get(self, name):
        return self._materials[self._strip(name)]['material']

    def _set(self, name, material):
        self._materials[self._strip(name)] = {'name': name, 'material': material}

    def _get_alias(self, name):
        return self._aliases[self._strip(name)]['refname']

    def _set_alias(self, name, refname):
        sname = self._strip(name)
        refname = self._strip(refname)
        self._aliases[sname] = {'name': name, 'refname': refname}

    def _resolve(self, material):
        if isinstance(material, str):
            material = self[material]
        if not isinstance(material, Material):
            raise ValueError("'material' must be a Material instance or a string")
        return material

    def _find_name_by_material(self, material):
        for nn, mat in self.items():
            if mat == material:
                return nn
        # raise ValueError(f"Broken material database. Material {material.name} not found.")


class MaterialsSubDatabase:
    """Class to handle a subset of the materials database for Xcoll."""

    def __init__(self, names, db, typename):
        self._names = names
        self._db = db
        self._typename = typename
        self._iter_index = 0
        self._iter_names = []

    def __getitem__(self, name):
        if name in self._names:
            refname = self._names[name]
            if refname not in self._db:
                raise KeyError(f"Broken database: Material with {self._typename} "
                               f"name '{name}' not found in the database.")
            return self._db[refname]
        raise KeyError(f"Material with name '{name}' not found in the database.")

    def __setitem__(self, name, material):
        if name in self._names:
            return  # Already in database
        material = self._db._resolve(material)
        # Check if material already exists (by identity, not by value).
        existing_name = self._db._find_name_by_material(material)
        if existing_name:
            self._names[name] = self._db._strip(existing_name)
        else:
            # If not, add new material
            if material.name is None:
                material.name = name        # This will add the material to the main database
            if self._typename == 'FLUKA' and material.fluka_name is None:
                material.fluka_name = name  # This will add the alias
            if self._typename == 'Geant4' and material.geant4_name is None:
                material.geant4_name = name # This will add the alias

    def __contains__(self, name):
        return name in self._names

    def __iter__(self):
        self._iter_index = 0
        self._iter_names = list(self._names.keys())
        return self

    def __next__(self):
        if self._iter_index >= len(self._iter_names):
            raise StopIteration
        value = self._names[self._iter_names[self._iter_index]]
        self._iter_index += 1
        return value

    def __len__(self):
        return len(self._names)

    def items(self):
        for name, mat in self._names.items():
            yield name, self._db[mat]

    def keys(self):
        for name in self._names.keys():
            yield name

    def values(self):
        for mat in self._names.values():
            yield self._db[mat]

    def rename_alias(self, name, new_name):
        if name not in self:
            raise KeyError(f"Alias '{name}' not found.")
        if new_name in self:
            if self[new_name] == self[name]:
                del self._names[name]
                return  # Already has this name
            raise ValueError(f"Material with name '{new_name}' already exists.")
        refname = self._names.pop(name)
        self._names[new_name] = refname

    def remove_alias(self, name):
        if name in self._names:
            del self._names[name]


db = MaterialsDatabase()

def _manually_add_material_to_db(mat, name, short_name=None, fluka_name=None, geant4_name=None):
    mat._name = name
    db._set(name, mat)
    if short_name:
        mat._short_name = short_name
        db._set_alias(short_name, name)
    if fluka_name:
        mat._fluka_name = fluka_name[:8]   # FLUKA material name max length is 8
        db.fluka._names[fluka_name] = db._strip(name)
    if geant4_name:
        mat._geant4_name = geant4_name
        db.geant4._names[geant4_name] = db._strip(name)



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025                  #
# ######################################### #

import io
import numpy as np


class BDSIMServer:

    def __init__(self):
        self.g4link = None


    def XtrackInterface(self, bdsimConfigFile=None, referencePdgId=None, referenceEk=None,
                        relativeEnergyCut=None, seed=None, batchMode=None, workdir=None):
        from g4interface import XtrackInterface
        self.g4link = XtrackInterface(bdsimConfigFile=bdsimConfigFile,
                                      referencePdgId=referencePdgId,
                                      referenceEk=referenceEk,
                                      relativeEnergyCut=relativeEnergyCut,
                                      seed=seed, batchMode=batchMode,
                                      workdir=workdir)


    def addCollimator(self, geant4_id, material, tipMaterial='', tipThickness=0, length=0,
                      apertureLeft=None, apertureRight=None, rotation=None, xOffset=0, yOffset=0,
                      side=None, jawTiltLeft=None, jawTiltRight=None, isACrystal=False):
        self.g4link.addCollimator(geant4_id, material, tipMaterial, tipThickness, length,
                                  apertureLeft=apertureLeft,
                                  apertureRight=apertureRight,
                                  rotation=rotation,
                                  xOffset=xOffset, yOffset=yOffset, side=side,
                                  jawTiltLeft=jawTiltLeft, jawTiltRight=jawTiltRight,
                                  isACrystal=isACrystal)


    def add_particles_and_collimate_return(self, blob, geant4_id, num_sent):
        buf = io.BytesIO(blob)
        loaded = np.load(buf)
        coords = [np.array(loaded[key]) for key in loaded.files]
        self.g4link.addParticles(coords)
        self.g4link.selectCollimator(geant4_id)
        self.g4link.collimate()
        products = self.g4link.collimateReturn(num_sent)
        # Pack results as a dict for serialization
        result_dict = {
            'x': products['x'],
            'xp': products['xp'],
            'y': products['y'],
            'yp': products['yp'],
            'zeta': products['zeta'],
            'p': products['p'],
            'q': products['q'],
            'm': products['m'],
            'weight': products['weight'],
            'pdg_id': products['pdg_id'],
            'parent_particle_id': products['parent_particle_id'],
            'state': products['state'],
            'n_hits': products['n_hits']
        }

        # Serialize to binary blob
        out_buf = io.BytesIO()
        np.savez(out_buf, **result_dict)
        return out_buf.getvalue()

    def clearData(self):
        self.g4link.clearData()



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from ..xoconstants import Constants, constant, group

class XcollParticleStates(Constants):
    _category_ = "particle_state" # auto-plural -> "particle_states"
    _reverse_  = "unique"         # builds particle_state_names
    _c_prefix_ = "XC"

    LOST_WITHOUT_SPEC       = constant(-300, "Lost in Xcoll but no specific cause recorded.")

    LOST_ON_EVEREST_BLOCK   = constant(-330, "Everest: absorbed by bulk material.")
    LOST_ON_EVEREST_COLL    = constant(-331, "Everest: collimator jaw absorption.")
    LOST_ON_EVEREST_CRYSTAL = constant(-332, "Everest: crystal absorption.")

    LOST_ON_FLUKA_BLOCK     = constant(-333, "FLUKA: absorbed by bulk material.")
    LOST_ON_FLUKA_COLL      = constant(-334, "FLUKA: collimator jaw absorption.")
    LOST_ON_FLUKA_CRYSTAL   = constant(-335, "FLUKA: crystal absorption.")

    LOST_ON_GEANT4_BLOCK    = constant(-336, "Geant4: absorbed by bulk material.")
    LOST_ON_GEANT4_COLL     = constant(-337, "Geant4: collimator jaw absorption.")
    LOST_ON_GEANT4_CRYSTAL  = constant(-338, "Geant4: crystal absorption.")

    LOST_ON_BLACK_ABSORBER  = constant(-340, "Lost on black absorber.")
    LOST_ON_BLACK_CRYSTAL   = constant(-341, "Lost on black crystal.")

    MASSLESS_OR_NEUTRAL     = constant(-350, "Massless or neutral particle.")
    ACC_IONISATION_LOSS     = constant(-351, "Not a real particle: Accumulated ionisation loss.")
    VIRTUAL_ENERGY          = constant(-352, "Not a real particle: Virtual energy deposition.")
    EXCITED_ION_STATE       = constant(-353, "An excited state of an ion (not supported by BDSIM or FLUKA).")

    ERR_INVALID_TRACK       = constant(-390, "Invalid track through Xcoll element.")
    ERR_NOT_IMPLEMENTED     = constant(-391, "Not implemented in Xcoll.")
    ERR_INVALID_XOFIELD     = constant(-392, "Invalid xofield in Xcoll element.")
    ERR                     = constant(-399, "Unknown Xcoll error.")

    HIT_ON_FLUKA_BLOCK      = constant(333, "Temporary variable to register hits. Should not be present in final states.")
    HIT_ON_FLUKA_COLL       = constant(334, "Temporary variable to register hits. Should not be present in final states.")
    HIT_ON_FLUKA_CRYSTAL    = constant(335, "Temporary variable to register hits. Should not be present in final states.")

    HIT_ON_GEANT4_BLOCK     = constant(336, "Temporary variable to register hits. Should not be present in final states.")
    HIT_ON_GEANT4_COLL      = constant(337, "Temporary variable to register hits. Should not be present in final states.")
    HIT_ON_GEANT4_CRYSTAL   = constant(338, "Temporary variable to register hits. Should not be present in final states.")

    # groups
    LOST_ON_BLOCK           = group(LOST_ON_EVEREST_BLOCK, LOST_ON_FLUKA_BLOCK, LOST_ON_GEANT4_BLOCK,
                                    info="Lost on any block")
    LOST_ON_COLL            = group(LOST_ON_EVEREST_COLL, LOST_ON_FLUKA_COLL, LOST_ON_GEANT4_COLL,
                                    info="Lost on any collimator")
    LOST_ON_CRYSTAL         = group(LOST_ON_EVEREST_CRYSTAL, LOST_ON_FLUKA_CRYSTAL, LOST_ON_GEANT4_CRYSTAL,
                                    info="Lost on any crystal")
    LOST_ON_ABSORBER        = group(LOST_ON_BLACK_ABSORBER, LOST_ON_BLACK_CRYSTAL,
                                    info="Lost on any absorber")
    LOST_AS_SPECIAL_STATE   = group(MASSLESS_OR_NEUTRAL, ACC_IONISATION_LOSS, VIRTUAL_ENERGY, EXCITED_ION_STATE,
                                    info="Special states or unsupported particles.")
    USE_IN_LOSSMAP          = group(LOST_ON_BLOCK, LOST_ON_COLL, LOST_ON_CRYSTAL, LOST_ON_ABSORBER, LOST_AS_SPECIAL_STATE,
                                    info="All states that should be used in loss maps.")



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

XCOLL_ELEMENTS_INIT_DEFAULTS = {
    # 'FlukaCollimator': {
    #     'length': 1., 'material': 'mogr'
    # },
    # 'FlukaCrystal': {
    #     'length': 1., 'material': 'si', 'bending_radius': 10., 'side': '+'
    # },
}



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from ...xoconstants import Constants, constant


class FlukaNames(Constants):
    _category_ = "fluka_name"
    _c_prefix_ = "XC_FLUKA"

    PHOTON_NAME                  = constant("PHOTON",    "Photon name as used in FLUKA.",                 pdg_id=22)

    ELECTRON_NAME                = constant("ELECTRON",  "Electron name as used in FLUKA.",               pdg_id=11)
    ELECTRON_NEUTRINO_NAME       = constant("NEUTRIE",   "Electron-neutrino name as used in FLUKA.",      pdg_id=12)
    MUON_NAME                    = constant("MUON-",     "Muon name as used in FLUKA.",                   pdg_id=13)
    MUON_NEUTRINO_NAME           = constant("NEUTRIM",   "Muon-neutrino name as used in FLUKA.",          pdg_id=14)
    TAU_NAME                     = constant("TAU-",      "Tau name as used in FLUKA.",                    pdg_id=15)
    TAU_NEUTRINO_NAME            = constant("NEUTRIT",   "Tau-neutrino name as used in FLUKA.",           pdg_id=16)

    POSITRON_NAME                = constant("POSITRON",  "Positron name as used in FLUKA.",               pdg_id=-11)
    ANTI_ELECTRON_NEUTRINO_NAME  = constant("ANEUTRIE",  "Anti-electron-neutrino name as used in FLUKA.", pdg_id=-12)
    ANTI_MUON_NAME               = constant("MUON+",     "Anti-muon name as used in FLUKA.",              pdg_id=-13)
    ANTI_MUON_NEUTRINO_NAME      = constant("ANEUTRIM",  "Anti-muon-neutrino name as used in FLUKA.",     pdg_id=-14)
    ANTI_TAU_NAME                = constant("TAU+",      "Anti-tau name as used in FLUKA.",               pdg_id=-15)
    ANTI_TAU_NEUTRINO_NAME       = constant("ANEUTRIT",  "Anti-tau-neutrino name as used in FLUKA.",      pdg_id=-16)

    PROTON_NAME                  = constant("PROTON",    "Proton name as used in FLUKA.",                 pdg_id=2212)
    ANTI_PROTON_NAME             = constant("APROTON",   "Anti-proton name as used in FLUKA.",            pdg_id=-2212)
    NEUTRON_NAME                 = constant("NEUTRON",   "Neutron name as used in FLUKA.",                pdg_id=2112)
    ANTI_NEUTRON_NAME            = constant("ANEUTRON",  "Anti-neutron name as used in FLUKA.",           pdg_id=-2112)
    TRITON_NAME                  = constant("TRITON",    "Triton name as used in FLUKA.",                 pdg_id=1000010030)
    DEUTERON_NAME                = constant("DEUTERON",  "Deuteron name as used in FLUKA.",               pdg_id=1000010020)
    HELIUM3_NAME                 = constant("3-HELIUM",  "He3 name as used in FLUKA.",                    pdg_id=1000020030)
    HELIUM4_NAME                 = constant("4-HELIUM",  "He4 name as used in FLUKA.",                    pdg_id=1000020040)

    PION_NAME                    = constant("PIZERO",    "π0 name as used in FLUKA.",                     pdg_id=111)
    PION_POS_NAME                = constant("PION+",     "π+ name as used in FLUKA.",                     pdg_id=211)
    PION_NEG_NAME                = constant("PION-",     "π- name as used in FLUKA.",                     pdg_id=-211)
    KAON_NAME                    = constant("KAONZERO",  "K0 name as used in FLUKA.",                     pdg_id=311)
    ANTI_KAON_NAME               = constant("AKAONZER",  "Anti-K0 name as used in FLUKA.",                pdg_id=-311)
    KAON_POS_NAME                = constant("KAON+",     "K+ name as used in FLUKA.",                     pdg_id=321)
    KAON_NEG_NAME                = constant("KAON-",     "K- name as used in FLUKA.",                     pdg_id=-321)
    KAON_LONG_NAME               = constant("KAONLONG",  "K long name as used in FLUKA.",                 pdg_id=130)
    KAON_SHORT_NAME              = constant("KAONSHRT",  "K short name as used in FLUKA.",                pdg_id=310)
    D_NAME                       = constant("D0",        "D0 name as used in FLUKA.",                     pdg_id=421)
    ANTI_D_NAME                  = constant("D0BAR",     "Anti-D0 name as used in FLUKA.",                pdg_id=-421)
    D_POS_NAME                   = constant("D+",        "D+ name as used in FLUKA.",                     pdg_id=411)
    D_NEG_NAME                   = constant("D-",        "D- name as used in FLUKA.",                     pdg_id=-411)
    D_S_POS_NAME                 = constant("DS+",       "Ds+ name as used in FLUKA.",                    pdg_id=431)
    D_S_NEG_NAME                 = constant("DS-",       "Ds- name as used in FLUKA.",                    pdg_id=-431)

    LAMBDA_NAME                  = constant("LAMBDA",    "Λ0 name as used in FLUKA.",                     pdg_id=3122)
    LAMBDA_C_POS_NAME            = constant("LAMBDAC+",  "Λc+ name as used in FLUKA.",                    pdg_id=4122)
    ANTI_LAMBDA_NAME             = constant("ALAMBDA",   "Anti-Λ0 name as used in FLUKA.",                pdg_id=-3122)
    ANTI_LAMBDA_C_NEG_NAME       = constant("ALAMBDC-",  "Anti-Λc- name as used in FLUKA.",               pdg_id=-4122)
    SIGMA_NAME                   = constant("SIGMAZER",  "Σ0 name as used in FLUKA.",                     pdg_id=3212)
    SIGMA_POS_NAME               = constant("SIGMA+",    "Σ+ name as used in FLUKA.",                     pdg_id=3222)
    SIGMA_NEG_NAME               = constant("SIGMA-",    "Σ- name as used in FLUKA.",                     pdg_id=3112)
    ANTI_SIGMA_NAME              = constant("ASIGMAZE",  "Anti-Σ0 name as used in FLUKA.",                pdg_id=-3212)
    ANTI_SIGMA_NEG_NAME          = constant("ASIGMA-",   "Anti-Σ- name as used in FLUKA.",                pdg_id=-3222)
    ANTI_SIGMA_POS_NAME          = constant("ASIGMA+",   "Anti-Σ+ name as used in FLUKA.",                pdg_id=-3112)
    XI_NAME                      = constant("XSIZERO",   "Ξ0 name as used in FLUKA.",                     pdg_id=3322)
    XI_NEG_NAME                  = constant("XSI-",      "Ξ- name as used in FLUKA.",                     pdg_id=3312)
    XI_C_NAME                    = constant("XSIC0",     "Ξc0 name as used in FLUKA.",                    pdg_id=4132)
    XI_C_POS_NAME                = constant("XSIC+",     "Ξc+ name as used in FLUKA.",                    pdg_id=4232)
    XI_C_PRIME_NAME              = constant("XSIPC0",    "Ξ'c0 name as used in FLUKA.",                   pdg_id=4312)
    XI_C_PRIME_POS_NAME          = constant("XSIPC+",    "Ξ'c+ name as used in FLUKA.",                   pdg_id=4322)
    ANTI_XI_NAME                 = constant("AXSIZERO",  "Anti-Ξ0 name as used in FLUKA.",                pdg_id=-3322)
    ANTI_XI_POS_NAME             = constant("AXSI+",     "Anti-Ξ+ name as used in FLUKA.",                pdg_id=-3312)
    ANTI_XI_C_NAME               = constant("AXSIC0",    "Anti-Ξc0 name as used in FLUKA.",               pdg_id=-4132)
    ANTI_XI_C_NEG_NAME           = constant("AXSIC-",    "Anti-Ξc- name as used in FLUKA.",               pdg_id=-4232)
    ANTI_XI_C_PRIME_NAME         = constant("AXSIPC0",   "Anti-Ξ'c0 name as used in FLUKA.",              pdg_id=-4312)
    ANTI_XI_C_PRIME_NEG_NAME     = constant("AXSIPC-",   "Anti-Ξ'c- name as used in FLUKA.",              pdg_id=-4322)
    OMEGA_NEG_NAME               = constant("OMEGA-",    "Ω- name as used in FLUKA.",                     pdg_id=3334)
    OMEGA_C_NAME                 = constant("OMEGAC0",   "Ωc0 name as used in FLUKA.",                    pdg_id=4332)
    ANTI_OMEGA_POS_NAME          = constant("AOMEGA+",   "Anti-Ω+ name as used in FLUKA.",                pdg_id=-3334)
    ANTI_OMEGA_C_NAME            = constant("AOMEGAC0",  "Anti-Ωc0 name as used in FLUKA.",               pdg_id=-4332)



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
from pathlib import Path
from numbers import Number

import xobjects as xo
import xtrack as xt
import xtrack.particles.pdg as pdg

from .geometry import XcollGeometry
from .physics_settings import PhysicsSettingsHelper
from ..materials import Material
from ..interaction_record import InteractionRecord
from ..compare import deep_equal
try:
    # TODO: once xaux is in Xsuite keep only this
    from xaux import FsPath, ranID
except (ImportError, ModuleNotFoundError):
    from ..xaux import FsPath, ranID


class BaseEngine(xo.HybridClass):
    _xofields = {
        '_particle_ref':      xt.Particles._XoStruct,
        '_seed':              xo.UInt64,
        '_capacity':          xo.Int64,
        '_relative_capacity': xo.Int64,
    }

    _int32 = False
    _only_protons = False
    _element_classes = None
    _uses_input_file = False
    _num_input_files = 1
    _uses_run_folder = False

    _depends_on = [Material, InteractionRecord, xt.RandomUniform,
                   xt.RandomExponential, xt.RandomNormal, xt.RandomRutherford,
                   xt.Drift, XcollGeometry]

    def __init__(self, **kwargs):
        if self._element_classes is None:
            raise NotImplementedError(f"{self.__class__.__name__} needs to define `_element_classes`!")
        # Initialise defaults
        self._cwd = None
        self._line = None
        self._masses = None
        self._verbose = False
        self._input_file = None
        self._environment = None
        self._element_dict = {}
        self._warning_given = False
        self._element_index = 0
        self._tracking_initialised = False
        self._deactivated_elements = {}
        kwargs.setdefault('_particle_ref', xt.Particles())
        kwargs.setdefault('_seed', 0)
        kwargs.setdefault('_capacity', 0)
        super().__init__(**kwargs)
        self._physics_settings = PhysicsSettingsHelper(self)

    def __del__(self, *args, **kwargs):
        self.stop()

    def _warn(self, error=None):
        if not self._warning_given:
            print(f"Warning: Failed to import {self.__class__.__name__} environment "
                + f"(did you compile?).\n{self.name.capitalize()} elements can be installed "
                + f"but are not trackable.", flush=True)
            self._warning_given = True
        self.stop()
        if error:
            raise error

    def _print(self, *args, **kwargs):
        if self.verbose:
            kwargs.setdefault('flush', True)
            print(*args, **kwargs)


    # ==================
    # === Properties ===
    # ==================

    @property
    def environment(self):
        if self._environment is None:
            raise RuntimeError(f"{self.__class__.__name__} environment not set up! "
                             + f"Do not manually create an instance of the engine, but use "
                             + f"xcoll.{self.__class__.__name__}.engine instead.")
        return self._environment

    @property
    def name(self):
        return self.__class__.__name__.replace('Engine', '').lower()

    @property
    def cwd(self):
        return self._cwd

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, val):
        self._verbose = val

    @property
    def line(self):
        return self._line

    @line.setter
    def line(self, val):
        if not val is None and not isinstance(val, xt.Line):
            self.stop()
            raise ValueError("`line` has to be an xt.Line object!")
        self._line = val

    @line.deleter
    def line(self):
        self._line = None

    @property
    def particle_ref(self):
        initial = xt.Particles().to_dict()
        current = self._particle_ref.to_dict()
        if deep_equal(initial, current):
            return None
        else:
            return self._particle_ref

    @particle_ref.setter
    def particle_ref(self, val):
        if val is None:
            self._particle_ref = xt.Particles()
        else:
            if isinstance(val, xt.line.LineParticleRef):
                val = val._resolved
            if not isinstance(val, xt.Particles):
                self.stop()
                raise ValueError("`particle_ref` has to be an xt.Particles object!")
            if val._capacity > 1:
                self.stop()
                raise ValueError("`particle_ref` has to be a single particle!")
            pdg_id = val.pdg_id[0]
            if pdg_id == 0:
                if self._only_protons:
                    pdg_id = pdg.get_pdg_id_from_name('proton')
                else:
                    self.stop()
                    raise ValueError(f"{self.__class__.__name__} allows the use of particles "
                                   + f"different than protons. Hence, `particle_ref` "
                                   + f"needs to have a valid pdg_id.")
            elif self._only_protons and pdg_id != pdg.get_pdg_id_from_name('proton'):
                self.stop()
                raise ValueError("{self.__class__.__name__} only supports protons!")
            self._particle_ref = val
            self._particle_ref.pdg_id[0] = pdg_id
        self._physics_settings.update()

    @particle_ref.deleter
    def particle_ref(self):
        self.particle_ref = None

    @property
    def capacity(self):
        if self._capacity == 0:
            return None
        else:
            return int(self._capacity)

    @capacity.setter
    def capacity(self, val):
        if val is None:
            val = 0
        if not isinstance(val, Number) or val < 0:
            self.stop()
            raise ValueError("`capacity` has to be a positive integer!")
        self._capacity = int(val)

    @capacity.deleter
    def capacity(self):
        self.capacity = None

    @property
    def relative_capacity(self):
        if self._relative_capacity == 0:
            return None
        else:
            return int(self._relative_capacity)

    @relative_capacity.setter
    def relative_capacity(self, val):
        if val is None:
            val = 2
        if not isinstance(val, Number) or val < 0:
            self.stop()
            raise ValueError("`relative_capacity` has to be a positive integer!")
        if val <= 1:
            self.stop()
            raise ValueError("`relative_capacity` has to be larger than 1!")
        self._relative_capacity = int(val)

    @relative_capacity.deleter
    def relative_capacity(self):
        self.relative_capacity = None

    @property
    def seed(self):
        if self._seed == 0:
            return None
        else:
            return self._seed

    @seed.setter
    def seed(self, val):
        if val is None:
            val = 0
        if not isinstance(val, Number) or val < 0:
            self.stop()
            raise ValueError("`seed` has to be a positive integer!")
        val = int(val)
        if self._int32:
            new_val = np.uint32(val)
        else:
            new_val = np.uint64(val)
        if new_val != val:
            self._print(f"Warning: type change for seed {val}. Using {new_val}.")
        self._seed = new_val

    @seed.deleter
    def seed(self):
        self.seed = None

    @property
    def input_file(self):
        if self._uses_input_file:
            return self._input_file

    @property
    def element_dict(self):
        return self._element_dict

    def __getattr__(self, name):
        if name != '_physics_settings' and hasattr(self, '_physics_settings') and name in self._physics_settings.all_flags:
            return getattr(self._physics_settings, name)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        if hasattr(self, '_physics_settings') and name in self._physics_settings.all_flags:
            return setattr(self._physics_settings, name, value)
        return super().__setattr__(name, value)

    # ======================
    # === Public Methods ===
    # ======================

    def physics_settings(self):
        return self._physics_settings.show()

    def start(self, *, clean=True, input_file=None, **kwargs):
        if not self.environment:
            self.stop()
            raise RuntimeError(f"{self.name.capitalize()} environment not set up! "
                             + f"Do not manually create an instance of the engine.")
        self.environment.assert_environment_ready()
        if self.is_running():
            self._print("Engine already running.")
            return

        # Clean up any leftover failed runs
        self.stop(clean=clean)

        print(f"Starting {self.__class__.__name__}...   ", flush=True, end='')
        if self.verbose:
            print("", flush=True)

        kwargs = self._pre_start(**kwargs)

        # This needs to be set in the ChildEngine, either in _start_engine() or at the start of tracking
        self._tracking_initialised = False

        # Resolve input file path before defining cwd
        if input_file:
            if not isinstance(input_file, (str,Path)):
                self.stop()
                raise ValueError("`input_file` has to be a string or Path!")
            if self._num_input_files > 1:
                input_file = [FsPath(f).expanduser().resolve() for f in input_file]
                for f in input_file:
                    if not f.exists():
                        self.stop()
                        raise ValueError(f"Input file {f} does not exist!")
            else:
                input_file = FsPath(input_file).expanduser().resolve()
                if not input_file.exists():
                    self.stop()
                    raise ValueError(f"Input file {input_file} does not exist!")

        # Set all engine properties that have a setter (this will remove these properties from the kwargs)
        self._starting_or_stopping = True # We need this to allow changing the element settings which otherwise are locked
        kwargs = self._set_engine_properties(**kwargs)

        # Do some preparations before creating the input file if needed
        kwargs = self._pre_input(**kwargs)

        # Create input file if needed (this will remove the kwargs relevant to the input file and physics)
        kwargs = self._use_input_file(input_file, **kwargs)
        if clean:
            self.clean_input_files(clean_all=False)
        self._starting_or_stopping = False

        # Start the engine in the ChildEngine
        self._start_engine(**kwargs)

        # Done starting
        if self.verbose:
            print(f"{self.__class__.__name__} started.", flush=True)
        else:
            print(f"Done.", flush=True)

    def stop(self, clean=False, **kwargs):
        kwargs = self._stop_engine(**kwargs)
        if clean:
            self.clean(clean_all=True, **kwargs)
        self._starting_or_stopping = True # We need this to allow changing the element settings which otherwise are locked
        self._restore_engine_properties(clean=clean)
        self._starting_or_stopping = False
        self._warning_given = False
        self._tracking_initialised = False
        self._environment.restore_environment()

    def is_running(self):
        if hasattr(self, '_starting_or_stopping') and self._starting_or_stopping:
            # We need this to allow changing the element settings which otherwise are locked
            return False
        # If we get here, we cannot say if the engine is running or not and we need an
        # implementation in the child class
        return self._is_running()


    def generate_input_file(self, *, clean=True, filename=None, **kwargs):
        '''This method manually generates an input file without starting the engine'''
        if not self._uses_input_file:
            self.stop()
            raise ValueError(f"{self.__class__.__name__} does not use input files!")
        if self._element_dict:
            self.stop()
            raise ValueError("Elements already assigned to engine (cannot regenerate input "
                           + "file after starting engine)!")

        # Set all engine properties that have a setter (this will remove these properties from the kwargs)
        cwd = kwargs.get('cwd')
        if cwd and FsPath(cwd).resolve() == FsPath.cwd().resolve():
            print("Warning: Cannot use current working directory as input folder. "
                + "Generating temporary folder instead.")
            kwargs.pop('cwd')
        self._starting_or_stopping = True # We need this to allow changing the element settings which otherwise are locked
        kwargs = self._set_engine_properties(**kwargs)

        # Do some preparations before creating the input file if needed
        kwargs = self._pre_input(**kwargs)

        # Create input file
        input_file, _ = self._generate_input_file(**kwargs)
        if not hasattr(input_file, '__iter__') or isinstance(input_file, str):
            # Some engines might create multiple input files (like Fluka)
            input_file = [input_file]

        # Move input file to desired location
        if filename is None:
            if input_file[0].parent != FsPath.cwd():
                new_input_file = [input_file[0].rename(Path.cwd() / input_file[0].name)]
            else:
                new_input_file = [input_file[0]]
        else:
            new_input_file = [input_file[0].rename(filename)]
        for file in input_file[1:]:
            if file.parent != new_input_file[0].parent:
                new_input_file.append(file.rename(new_input_file[0].parent / file.name))
            else:
                new_input_file.append(file)

        # Clean up
        if clean:
            if input_file[0].parent == new_input_file[0].parent:
                self.clean_input_files(clean_all=False, input_file=input_file)
            else:
                self.clean_input_files(clean_all=True, input_file=input_file)
        self._restore_engine_properties(clean=clean)
        self._starting_or_stopping = False

        return new_input_file[0] if self._num_input_files==1 else new_input_file


    def assert_particle_ref(self):
        if self.particle_ref is None:
            self.stop()
            raise ValueError(f"{self.__class__.__name__} reference particle not set!")

    def assert_ready_to_track_or_skip(self, coll, particles, _necessary_attributes=[], keep_p0c_constant=True):
        '''Asserts that the engine is ready to track the given element and particles.'''
        self._assert_element(coll)

        missing_attributes = False
        for attr in _necessary_attributes:
            if not hasattr(coll, attr) or not getattr(coll, attr):
                missing_attributes = True

        if not coll.active or not coll._tracking or not coll.jaw or missing_attributes:
            return False

        npart = particles._num_active_particles
        if npart == 0:
            return False
        if not isinstance(particles._buffer.context, xo.ContextCpu):
            self.stop()
            raise ValueError(f"{self.__class__.__name__} only supports CPU contexts!")

        assert self.environment.compiled
        if not self.is_running():
            self.stop()
            raise ValueError(f"{self.__class__.__name__} not yet running!\nPlease do this "
                           + f"first, by calling xcoll.{self.__class__.__name__}.start().")

        self.assert_particle_ref()
        if abs(particles.mass0 - self.particle_ref.mass0) > 1e-3:
            # This should only happen at the start of the tracking.
            pdg_id = self.particle_ref.pdg_id[0]
            if self._masses is not None and abs(pdg_id) in self._masses:
                # The reference particle in the engine was updated to match the scattering code.
                # Match the particles object to it.
                new_mass = self.particle_ref.mass0
                old_energy0 = particles.energy0[0]
                particles.mass0 = new_mass
                if keep_p0c_constant:
                    particles._update_refs(p0c=particles.p0c[0])
                else:
                    particles._update_refs(energy0=old_energy0)
                assert np.isclose(particles.energy0[0]**2, particles.p0c[0]**2 + particles.mass0**2)
                assert np.isclose(particles.mass0, new_mass)
                # No need to update chi, delta, etc as they depend on the RATIO m0/m and this is unchanged.
                self._print(f"Warning: reference mass in particles object differs "
                          + f"from reference mass in engine. Overwritten by the latter.")
            else:
                # The reference particle in the engine was changed unintentionally
                self.stop()
                raise ValueError(f"Error in reference mass of `particles`: not in sync with "
                            + f"{self.name} reference particle!\nRebuild the particles object "
                            + f"using the {self.__class__.__name__} reference particle.")
        if abs(particles.q0 - self.particle_ref.q0) > 1e-3:
            self.stop()
            raise ValueError(f"Error in reference charge of `particles`: not in sync with "
                           + f"{self.name} reference particle!\nRebuild the particles object "
                           + f"using the {self.__class__.__name__} reference particle.")
        if not self._only_protons:
            if np.any([pdg_id == 0 for pdg_id in particles.pdg_id]):
                self.stop()
                raise ValueError("Some particles are missing the pdg_id!")
            if particles._num_active_particles + particles._num_lost_particles == particles._capacity \
            and not np.any(particles.particle_id != particles.parent_particle_id):
                # Only raise this error at the start, e.g. when no secondaries are present yet.
                # It will get caught later during tracking, which will provide a more logical error.
                self.stop()
                raise ValueError("Particles capacity equal to size! Please provide extra capacity "
                               + "for secondaries.")
        return True


    def clean(self, **kwargs):
        '''Cleans all input and output files generated by the engine.
        Use clean_all (default: False) to also delete generated input files.'''
        self.clean_input_files(**kwargs)
        self.clean_output_files(**kwargs)

    def clean_input_files(self, clean_all=False, **kwargs):
        kwargs = self._get_input_cwd_for_cleaning(**kwargs)
        files_to_delete = self._get_input_files_to_clean(**kwargs)
        if not clean_all:
            files_to_delete = [f for f in files_to_delete
                    if f not in self._all_input_files(kwargs['input_file'])]
        for f in files_to_delete:
            if f is not None and f.exists():
                if f.is_dir():
                    f.rmtree()
                else:
                    f.unlink()

    def clean_output_files(self, clean_all=False, **kwargs):
        kwargs = self._get_input_cwd_for_cleaning(**kwargs)
        files_to_delete = self._get_output_files_to_clean(**kwargs)
        if not clean_all:
            files_to_delete = [f for f in files_to_delete
                    if f not in self._all_input_files(kwargs['input_file'])]
        for f in files_to_delete:
            if f is not None and f.exists():
                if f.is_dir():
                    f.rmtree()
                else:
                    f.unlink()


    # =======================
    # === Private Methods ===
    # =======================

    # For all the engine fields, they can either be set in advance on the engine,
    # or they can be set when the engine is started. In the latter case, the values
    # are temporary and the original will be restored when the engine is stopped.

    def _set_property(self, prop, kwargs):
        val = kwargs.pop(prop, None)
        if val is not None:
            # We only need to update the property when it is not None
            setattr(self, f'_old_{prop}', getattr(self, prop))
            setattr(self, prop, val)

    def _set_engine_properties(self, **kwargs):
        # We need to set the following properties first as they are needed by the others
        self._set_property('verbose', kwargs)
        self._set_property('line', kwargs)
        # The following properties have a specific logic
        self._use_seed(kwargs.pop('seed', None))
        self._use_particle_ref(kwargs.pop('particle_ref', None))
        self._sync_line_particle_ref()
        self._get_elements(kwargs.pop('elements', None), kwargs.pop('names', None))
        self._set_cwd(kwargs.pop('cwd', None))
        # Now we can set the rest of the properties
        self._set_property('capacity', kwargs)
        self._set_property('relative_capacity', kwargs)
        for ff in self._physics_settings.all_flags:
            self._set_property(ff, kwargs)
        return kwargs

    def _restore_engine_properties(self, clean=False):
        # Reset particle_ref in the line
        if hasattr(self, '_old_line_particle_ref'):
            self.line.particle_ref = self._old_line_particle_ref
            del self._old_line_particle_ref
        # The following properties have a specific logic
        self._reactivate_elements()
        self._reset_cwd(clean=clean)
        # Reset all other properties
        self_attributes = self.__dict__.copy()
        for kk, vv in self_attributes.items():
            if kk.startswith('_old_'):
                prop = kk[5:]
                setattr(self, prop, vv)
                delattr(self, kk)
        self._input_file = None
        self._element_dict = {}

    def _use_seed(self, seed=None):
        if seed is None:
            if self.seed is None:
                rng = np.random.default_rng()
                if self._int32:
                    self.seed = rng.integers(0, int(2**31)) # 32-bit signed
                else:
                    self.seed = rng.integers(0, int(2**63)) # 64-bit signed
        else:
            self._old_seed = self.seed
            self.seed = seed
        self._print(f"Using seed {self.seed}.")

    def _use_particle_ref(self, particle_ref=None, keep_p0c_constant=True):
        # Prefer: provided particle_ref > existing particle_ref > particle_ref from line
        if particle_ref is not None:
            self._old_particle_ref = self.particle_ref
            self.particle_ref = particle_ref
        elif self.particle_ref is None:
            if self.line is None or not hasattr(self.line, 'particle_ref') \
            or self.line.particle_ref is None:
                self.stop()
                raise ValueError("Need to provide either a line with a reference "
                               + "particle, or `particle_ref`.")
            self._old_particle_ref = self.particle_ref
            self.particle_ref = self.line.particle_ref
        self._print(f"Using {pdg.get_name_from_pdg_id(self.particle_ref.pdg_id[0])} "
                  + f"with momentum {self.particle_ref.p0c[0]/1.e9:.1f} GeV.")
        if self._masses is not None:
            mass = self.particle_ref.mass0
            pdg_id = self.particle_ref.pdg_id[0]
            if abs(pdg_id) in self._masses:
                new_mass = self._masses[abs(pdg_id)]
                if abs(mass-new_mass)/mass > 1.e-12:
                    old_energy0 = self.particle_ref.energy0[0]
                    self.particle_ref.mass0  = new_mass
                    if keep_p0c_constant:
                        self.particle_ref._update_refs(p0c=self.particle_ref.p0c[0])
                    else:
                        self.particle_ref._update_refs(energy0=old_energy0)
                    assert np.isclose(self.particle_ref.energy0[0]**2, self.particle_ref.p0c[0]**2 + self.particle_ref.mass0**2)
                    assert np.isclose(self.particle_ref.mass0, new_mass)
                    self._print(f"Warning: given mass of {mass} eV for "
                            + f"{pdg.get_name_from_pdg_id(pdg_id)} differs from {self.name} "
                            + f"mass of {new_mass} eV.\nReference particle mass is "
                            + f"overwritten by the latter.")
            else:
                self._print(f"Warning: No {self.name} reference mass known for particle "
                        + f"{pdg.get_name_from_pdg_id(pdg_id)}!\nIf the reference mass "
                        + f"provided ({mass} eV) differs from the one used internally "
                        + f"by {self.name}, differences in energy might be observed.\nOnce "
                        + f"the {self.name} reference mass is known, contact the devs to "
                        + f"input it in the code.")

    def _sync_line_particle_ref(self):
        if self.line is None:
            return
        if self.line.particle_ref is not None \
        and not deep_equal(self.line.particle_ref.to_dict(),
                                     self.particle_ref.to_dict()):
            self._print("Found different reference particle in line. Temporarily overwritten.")
            val = self.line.particle_ref
            if isinstance(val, xt.line.LineParticleRef):
                val = val._resolved
            self._old_line_particle_ref = val
            self.line.particle_ref = self.particle_ref

    def _get_new_element_name(self):
        name = f"{self.name}_el_{self._element_index}"
        self._element_index += 1
        return name

    def _assert_element(self, element):
        if not isinstance(element, self._element_classes):
            self.stop()
            raise ValueError(f"Element {element.name} is not a "
                            + ", or a ".join([c.__name__ for c in self._element_classes])
                            + ".")

    def _get_elements(self, elements=None, names=None):
        if elements is not None and (not hasattr(elements, '__iter__') or isinstance(elements, str)):
            elements = [elements]
        if names is not None and (not hasattr(names, '__iter__') or isinstance(names, str)):
            names = [names]
        if self.line is None:
            if elements is None:
                self.stop()
                raise ValueError("Need to provide either `line` or `elements`.")
            if names is None:
                names = []
                for ee in elements:
                    if hasattr(ee, 'name') and ee.name:
                        names.append(ee.name)
                    else:
                        name = self._get_new_element_name()
                        names.append(name)
                        ee.name = name
            elif len(names) == len(elements):
                for ee, name in zip(elements, names):
                    if hasattr(ee, 'name'):
                        if ee.name != name:
                            self._print(f"Warning: Element name {ee.name} changed to {name}.")
                            ee.name = name
            else:
                self.stop()
                raise ValueError("Length of `elements` and `names` doesn't match.")
        else:
            if elements is not None:
                self.stop()
                raise ValueError("Cannot provide both `line` and `elements`.")
            if names is None:
                elements, names = self.line.get_elements_of_type(self._element_classes)
                # This method can return duplicate elements if there are subclasses (like Geant4Collimator
                # and Geant4CollimatorTip) as elements of the child type will also be found by searching
                # for the parent type.
                d = {}
                for nn, ee in zip(names, elements):
                    d[nn] = ee # last one wins, not important here
                names, elements = list(d.keys()), list(d.values())
            else:
                elements = [self.line[name] for name in names]
        this_names = []
        this_elements = []
        for ee, name in zip(elements, names):
            self._assert_element(ee)
            if ee.jaw is None:
                self._print(f"Warning: Jaw not set for {name}. Ignoring.")
                self._deactivate_element(ee)
            elif not ee.active:
                self._print(f"Warning: Element {name} is not active. Ignoring.")
                self._deactivate_element(ee)
            else:
                this_names.append(name)
                this_elements.append(ee)
        if len(this_elements) == 0:
            self.stop()
            raise ValueError(f"No active {self.name} elements found!")
        if len(set(this_names)) != len(this_names):
            self.stop()
            raise ValueError(f"Duplicate names found in {self.name} elements: {this_names}. "
                           + f"Please provide unique names for each element.")
        self._element_dict = dict(zip(this_names, this_elements))

    def _deactivate_element(self, el):
        self._deactivated_elements[el.name] = [el, el.active or True]
        if hasattr(el, 'active'):
            el.active = False
        self._remove_element(el)

    def _reactivate_elements(self):
        names = list(self._deactivated_elements.keys())
        for name in names:
            ee, was_active = self._deactivated_elements.pop(name)
            self._restore_element(ee)
            if hasattr(ee, 'active'):
                ee.active = was_active

    def _set_cwd(self, cwd=None):
        if self._uses_run_folder:
            if cwd is not None:
                cwd = FsPath(cwd).expanduser().resolve()
                if cwd.exists() and cwd != FsPath.cwd():
                    # Check if the folder already exists
                    # If the specified folder is the current one, we do not need to rename
                    i = 0
                    while (cwd.parent / f'{cwd.name}_{i:0>4}').exists():
                        i += 1
                        if i > 9999:
                            self.stop()
                            raise ValueError(f"Too many folders with the same "
                                           + f"name {cwd}!")
                    cwd = cwd.parent / f'{cwd.name}_{i:0>4}'
            else:
                ran_str = ranID(only_alphanumeric=True)
                cwd = FsPath.cwd() / f'{self.name}_run_{ran_str}'
            self._cwd = cwd
            cwd.mkdir(parents=True, exist_ok=True)
            # self._old_cwd = FsPath.cwd()
            # os.chdir(cwd)

    def _reset_cwd(self, clean=False):
        if clean and self.cwd is not None and self.cwd != FsPath.cwd():
            try:
                # Only works if the folder is empty
                self.cwd.rmdir()
            except:
                self._print(f"Warning: Failed to remove temporary folder "
                            + f"{self.cwd}.")
        self._cwd = None

    def _use_input_file(self, input_file=None, **kwargs):
        if self._uses_input_file:
            if input_file is None:
                input_file, kwargs = self._generate_input_file(**kwargs)
            if not hasattr(input_file, '__iter__') or isinstance(input_file, (str,Path)):
                # Some engines might need multiple input files (like Fluka)
                input_file = [input_file]
            input_file = [FsPath(f).expanduser().resolve() for f in input_file]
            new_files = []
            for file in input_file:
                if not file.exists():
                    self.stop()
                    raise ValueError(f"Input file {file.as_posix()} not found!")
                if file.parent != self.cwd and self._uses_run_folder:
                    if self.cwd is None:
                        self.stop()
                        raise ValueError("Cannot copy input file to working directory: "
                                       + "working directory not set!")
                    file.copy_to(self.cwd, method='mount')
                    new_files.append(self.cwd / file.name)
                else:
                    new_files.append(file)
            self._input_file = new_files[0] if self._num_input_files==1 else new_files
            self._match_input_file()
        return kwargs

    def _get_input_cwd_for_cleaning(self, **kwargs):
        # Get the input file and the cwd
        input_file = kwargs.get('input_file', None)
        cwd = kwargs.get('cwd', None)
        if input_file is None:
            if self.input_file is not None:
                input_file = self.input_file
            if cwd is None and self.cwd is not None:
                cwd = self.cwd
        else:
            if not hasattr(input_file, '__iter__') or isinstance(input_file, (str,Path)):
                input_file = [input_file]
            input_file = [FsPath(ff) for ff in input_file]
            if cwd is None:
                cwd = input_file[0].parent
            if len(input_file) == 1:
                input_file = input_file[0]
        kwargs['input_file'] = input_file
        if self._uses_run_folder:
            kwargs['cwd'] = cwd
        else:
            kwargs['cwd'] = FsPath.cwd()
        return kwargs


    def _mask_particle_return_types(self, pdg_id, q_new):
        if self.return_all:
            # Allow everything and exclude
            mask_new = np.ones_like(pdg_id, dtype=bool)
        else:
            # Allow nothing and include
            mask_new = np.zeros_like(pdg_id, dtype=bool)

        # General categories
        mask_new[pdg_id > 1000000000] = self.return_ions
        # PDG ID of mesons: from .*0XX. where X != 0 and . is any digit
        mask_new[(pdg_id > 0) & (pdg_id // 10 % 10 != 0) & (pdg_id // 100 % 10 != 0)
                              & (pdg_id // 1000 % 10 == 0)] = self.return_other_mesons
        mask_new[(pdg_id < 0) & (-pdg_id // 10 % 10 != 0) & (-pdg_id // 100 % 10 != 0)
                              & (-pdg_id // 1000 % 10 == 0)] = self.return_other_mesons
        # PDG ID of baryons: from XXX. where X != 0 and . is any digit
        mask_new[(pdg_id > 1000) & (pdg_id < 9000) & (pdg_id // 10 % 10 != 0) & (pdg_id // 100 % 10 != 0)
                                 & (pdg_id // 1000 % 10 != 0)] = self.return_other_baryons    # PDG ID of from XX0X is a diquark
        mask_new[(pdg_id < -1000) & (pdg_id > -9000) & (-pdg_id // 10 % 10 != 0) & (-pdg_id // 100 % 10 != 0)
                                  & (-pdg_id // 1000 % 10 != 0)] = self.return_other_baryons

        if not self.return_neutral:
            # General modifier, has to be before more specific return types,
            # as other neutral particles might have been specifically activated.
            mask_new[np.abs(q_new) < 1.e-12] = False

        mask_new[pdg_id == 22] = self.return_photons
        mask_new[(pdg_id == 11) | (pdg_id == -11)] = self.return_electrons
        mask_new[(pdg_id == 12) | (pdg_id == -12)] = self.return_electrons and self.return_neutrinos
        mask_new[(pdg_id == 13) | (pdg_id == -13)] = self.return_muons
        mask_new[(pdg_id == 14) | (pdg_id == -14)] = self.return_muons and self.return_neutrinos
        mask_new[(pdg_id == 15) | (pdg_id == -15)] = self.return_tauons
        mask_new[(pdg_id == 16) | (pdg_id == -16)] = self.return_tauons and self.return_neutrinos
        mask_new[(pdg_id == 211) | (pdg_id == -211)] = self.return_pions
        mask_new[(pdg_id == 111)] = self.return_pions and self.return_neutral
        mask_new[(pdg_id == 321) | (pdg_id == -321)] = self.return_kaons
        mask_new[(pdg_id == 130) | (pdg_id == -130)] = self.return_kaons and self.return_neutral
        mask_new[(pdg_id == 310) | (pdg_id == -310)] = self.return_kaons and self.return_neutral
        mask_new[(pdg_id == 311) | (pdg_id == -311)] = self.return_kaons and self.return_neutral
        mask_new[(pdg_id == 2212) | (pdg_id == -2212)] = self.return_protons
        mask_new[(pdg_id == 2112) | (pdg_id == -2112)] = self.return_neutrons
        return mask_new


    # =================================================
    # === Methods to be overwritten by child engine ===
    # =================================================

    def _start_engine(self, **kwargs):
        raise NotImplementedError(f"Need to implement `_start_engine` for "
                                   + f"{self.__class__.__name__}!")

    def _stop_engine(self, **kwargs):
        raise NotImplementedError(f"Need to implement `_stop_engine` for "
                                   + f"{self.__class__.__name__}!")

    def _is_running(self, **kwargs):
        raise NotImplementedError(f"Need to implement `_is_running` for "
                                   + f"{self.__class__.__name__}!")

    def _match_input_file(self, **kwargs):
        if self._uses_input_file:
            raise NotImplementedError(f"Need to implement `_match_input_file` for "
                                   + f"{self.__class__.__name__}!")

    def _generate_input_file(self, **kwargs):
        if self._uses_input_file:
            raise NotImplementedError("Need to implement `_generate_input_file` for "
                                   + f"{self.__class__.__name__}!")


    # ============================================================
    # === Methods to be optionally overwritten by child engine ===
    # ============================================================

    def _get_input_files_to_clean(self, input_file=None, cwd=None, clean_all=False, **kwargs):
        return []

    def _get_output_files_to_clean(self, input_file=None, cwd=None, clean_all=False, **kwargs):
        return []

    def _all_input_files(self, input_file=None):
        if self._uses_input_file:
            if self._num_input_files == 1:
                return [self._input_file]
            else:
                raise NotImplementedError(f"Need to implement `_all_input_files` for "
                                    + f"{self.__class__.__name__}!")
        else:
            return []

    def _remove_element(self, el):
        pass

    def _restore_element(self, el):
        pass

    def _pre_start(self, **kwargs):
        return kwargs

    def _pre_input(self, **kwargs):
        # Do some preparations before creating the input file if needed
        return kwargs



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xtrack as xt
from ...interaction_record import InteractionRecord
from ...general import _pkg_root

class XcollGeometry(xt.BeamElement):
    _xofields = {}

    allow_track = False

    _depends_on = [xt.Drift, xt.XYShift, xt.SRotation, xt.YRotation, InteractionRecord]

    _extra_c_sources = [
        _pkg_root.joinpath('scattering_routines','geometry','sort.h'),
        _pkg_root.joinpath('scattering_routines','geometry','segments.h'),
        _pkg_root.joinpath('scattering_routines','geometry','objects.h'),
        _pkg_root.joinpath('scattering_routines','geometry','methods.h'),
        _pkg_root.joinpath('scattering_routines','geometry','get_s.h'),
        _pkg_root.joinpath('scattering_routines','geometry','rotation.h'),
        _pkg_root.joinpath('scattering_routines','geometry','collimator_geometry.h'),
        _pkg_root.joinpath('scattering_routines','geometry','crystal_geometry.h')
    ]



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025                  #
# ######################################### #

import gc
import numpy as np
from numbers import Number

import xobjects as xo

from .rpyc import launch_rpyc_with_port # remove after geant4 bugfix
from .bdsim_config import create_bdsim_config_file, get_collimators_from_input_file
from .reference_masses import geant4_masses_src
from ..engine import BaseEngine
from ...general import _pkg_root


class Geant4Engine(BaseEngine):

    _xofields = BaseEngine._xofields | {
        '_relative_energy_cut':        xo.Float64
        # 'random_freeze_state':         xo.Int64,  # to be implemented; number of randoms already sampled, such that this can be taken up again later
    }

    _int32 = True
    _uses_input_file = True
    _uses_run_folder = True

    _depends_on = [BaseEngine]

    _extra_c_sources = [geant4_masses_src]

    def __init__(self, **kwargs):
        # Set element classes dynamically
        from ...beam_elements import Geant4Collimator, Geant4CollimatorTip, Geant4Crystal
        self.__class__._element_classes = (Geant4Collimator, Geant4CollimatorTip, Geant4Crystal)
        # Initialise geant4-only defaults
        self._g4link = None
        self._server = None # remove after geant4 bugfix
        self._conn = None # remove after geant4 bugfix
        super().__init__(**kwargs)
        self._already_started = False
        # Set default values for new properties
        self.relative_energy_cut = None
        self.reentry_protection_enabled = None
        self._physics_settings._use_cuts = False


    # ======================
    # === New Properties ===
    # ======================

    @property
    def relative_energy_cut(self):
        return self._relative_energy_cut

    @relative_energy_cut.setter
    def relative_energy_cut(self, val):
        if val is None:
            val = 0.1
        if not isinstance(val, Number) or val <= 0:
            self.stop()
            raise ValueError("`relative_energy_cut` has to be a strictly postive number!")
        self._relative_energy_cut = val

    @property
    def reentry_protection_enabled(self):
        return self._reentry_protection_enabled

    @reentry_protection_enabled.setter
    def reentry_protection_enabled(self, val):
        if val is False:
            print("Warning: Disabling re-entry protection can lead to crashes!")
            print("         Only disable if you know what you are doing.")
        elif val is None:
            try:
                import rpyc
            except ImportError as e:
                val = False
            else:
                val = True
        elif not isinstance(val, bool):
            self.stop()
            raise ValueError("`reentry_protection_enabled` has to be a boolean!")
        self._reentry_protection_enabled = val

    # ============================
    # === Overwrite Properties ===
    # ============================

    @property
    def capacity(self):
        return None  # Geant4 capacity is dynamic

    @property
    def relative_capacity(self):
        return None  # Geant4 capacity is dynamic

    # =================================
    # === Base methods to overwrite ===
    # =================================

    def _set_engine_properties(self, **kwargs):
        kwargs = super()._set_engine_properties(**kwargs)
        self._set_property('relative_energy_cut', kwargs)
        self._set_property('reentry_protection_enabled', kwargs)
        return kwargs

    def _pre_input(self, **kwargs):
        coll_id = 1
        for el in self._element_dict.values():
            el.geant4_id = f'XcollG4.{coll_id}'  # TODO: will be provided by new BDSIM interface
            coll_id += 1
        return kwargs

    def _generate_input_file(self, **kwargs):
        input_file, kwargs = create_bdsim_config_file(element_dict=self._element_dict,
                                particle_ref=self.particle_ref, verbose=self.verbose,
                                cwd=self.cwd, **kwargs)
        # The only thing left in kwargs are parameters to start the engine
        return input_file, kwargs

    def _start_engine(self, **kwargs):
        from ...beam_elements import BaseCrystal, Geant4CollimatorTip

        Ekin = self.particle_ref.energy0[0] - self.particle_ref.mass0
        try:
            from g4interface import XtrackInterface
        except (ModuleNotFoundError, ImportError) as error:
            self.stop(clean=True)
            self._warn(error)
            return

        if self.reentry_protection_enabled:
            ### remove this part after geant4 bug fixed
            try:
                import rpyc
            except (ModuleNotFoundError, ImportError) as e:
                self.stop()
                raise ImportError("Failed to import rpyc. Cannot connect to BDSIM.") from e
            self._server, port = launch_rpyc_with_port(log_path=self.cwd / "rpyc.log")
            self._conn = rpyc.classic.connect('localhost', port=port)
            self._conn._config['sync_request_timeout'] = 1240 # Set timeout to 1240 seconds
            self._conn.execute('import sys')
            self._conn.execute(f'sys.path.append("{self._environment.data_dir.as_posix()}")')  # For g4interface on server
            self._conn.execute(f'sys.path.append("{(_pkg_root / "scattering_routines" / "geant4").as_posix()}")')
            self._conn.execute('import engine_server')
            self._g4link = self._conn.namespace['engine_server'].BDSIMServer()
            self._g4link.XtrackInterface(bdsimConfigFile=self.input_file.as_posix(),
                                         referencePdgId=self.particle_ref.pdg_id[0],
                                         referenceEk=Ekin,
                                         relativeEnergyCut=self.relative_energy_cut,
                                         seed=self.seed, batchMode=True,
                                         workdir=self.cwd.as_posix())
        else:
            if self._already_started:
                self.stop(clean=True)
                raise RuntimeError("Cannot restart Geant4 engine in non-reentry-safe mode. "
                                 + "Please exit this Python process. Do pip install rpyc "
                                 + "to avoid this limitation.")

            self._g4link = XtrackInterface(bdsimConfigFile=self.input_file.as_posix(),
                                           referencePdgId=self.particle_ref.pdg_id[0],
                                           referenceEk=Ekin,
                                           relativeEnergyCut=self.relative_energy_cut,
                                           seed=self.seed, batchMode=True,
                                           workdir=self.cwd.as_posix())

        for el in self._element_dict.values():
            side = 2 if el._side == -1 else el._side
            jaw_L = 0.1 if el.jaw_L is None else el.jaw_L
            jaw_R = -0.1 if el.jaw_R is None else el.jaw_R
            tilt_L = 0.0 if el.tilt_L is None else el.tilt_L
            tilt_R = 0.0 if el.tilt_R is None else el.tilt_R
            jaw_L -= 1.e-9  # Correct for 1e-9 shift that is added in BDSIM
            jaw_R += 1.e-9  # Correct for 1e-9 shift that is added in BDSIM
            if jaw_L < jaw_R:
                self.stop(clean=True)
                raise ValueError(f"Geant4Collimator {el.name} has jaw_L < jaw_R: "
                             + f"jaw_L={el.jaw_L}, jaw_R={el.jaw_R}!")
            xOffset = 0
            yOffset = 0
            if jaw_L <= 0:
                xOffset = (jaw_L + jaw_R)/2
                jaw_L -= xOffset
                jaw_R -= xOffset
                xOffset_temp = xOffset*np.cos(np.deg2rad(el.angle_L)) - yOffset*np.sin(np.deg2rad(el.angle_L))
                yOffset = xOffset*np.sin(np.deg2rad(el.angle_L)) + yOffset*np.cos(np.deg2rad(el.angle_L))
                xOffset = xOffset_temp
            if jaw_R >= 0:
                xOffset = (jaw_L + jaw_R)/2
                jaw_L -= xOffset
                jaw_R -= xOffset
                xOffset_temp = xOffset*np.cos(np.deg2rad(el.angle_R)) - yOffset*np.sin(np.deg2rad(el.angle_R))
                yOffset = xOffset*np.sin(np.deg2rad(el.angle_R)) + yOffset*np.cos(np.deg2rad(el.angle_R))
                xOffset = xOffset_temp
            tip_material = el.tip_material.geant4_name if isinstance(el, Geant4CollimatorTip) else ''
            tip_thickness = el.tip_thickness if isinstance(el, Geant4CollimatorTip) else 0
            self._g4link.addCollimator(f'{el.geant4_id}', el.material.geant4_name,
                                       tip_material, tip_thickness, el.length,
                                       apertureLeft=jaw_L, apertureRight=-jaw_R,
                                       rotation=np.deg2rad(el.angle),
                                       xOffset=xOffset, yOffset=yOffset, side=side,
                                       jawTiltLeft=tilt_L, jawTiltRight=tilt_R,
                                       isACrystal=isinstance(el, BaseCrystal))
        self._already_started = True

    def _stop_engine(self, **kwargs):
        del self._g4link
        gc.collect()
        self._g4link = None
        if self.reentry_protection_enabled and self._server: # remove after geant4 bugfix
            self._server.terminate() # remove after geant4 bugfix
            self._server = None # remove after geant4 bugfix
        return kwargs

    def _is_running(self):
        return self._g4link is not None

    def _get_input_files_to_clean(self, input_file, cwd, **kwargs):
        if cwd is None or input_file is None:
            return []
        return [cwd / input_file]

    def _get_output_files_to_clean(self, input_file, cwd, **kwargs):
        if cwd is None:
            return []
        files_to_delete = ['rpyc.log', 'geant4.out', 'geant4.err',
                           'engine.out', 'engine.err', 'root.out',
                           'root.err']
        return [cwd / f for f in files_to_delete]

    def _match_input_file(self):
        # Read the elements in the input file and compare to the elements in the engine,
        # overwriting parameters where necessary
        input_dict = get_collimators_from_input_file(self.input_file)
        for name in input_dict:
            if name not in self._element_dict:
                self.stop()
                raise ValueError(f"Element {name} in input file not found in engine!")
        for name, ee in self._element_dict.items():
            from ...beam_elements import Geant4CollimatorTip
            if name not in input_dict:
                self._print(f"Warning: Geant4Collimator {name} not in Geant4 input file! "
                          + f"Maybe it was fully open. Deactivated")
                self._deactivate_element(ee)
                continue
            self._assert_element(ee)
            ee.geant4_id = input_dict[name]['geant4_id']
            if not np.isclose(ee.length, input_dict[name]['length'], atol=1e-9):
                self._print(f"Warning: Length of {name} differs from input file "
                        + f"({ee.length} vs {input_dict[name]['length']})! Overwritten.")
                ee.length = input_dict[name]['length']
            if not np.isclose(ee.angle, input_dict[name]['angle'], atol=1e-9):
                self._print(f"Warning: Angle of {name} differs from input file "
                        + f"({ee.angle} vs {input_dict[name]['angle']})! Overwritten.")
                ee.angle = input_dict[name]['angle']
            if ee.material.geant4_name != input_dict[name]['material'] \
            and ee.material.name != input_dict[name]['material']:
                self.stop()
                raise ValueError(f"Material of {name} differs from input file "
                            + f"({ee.material.geant4_name or ee.material.name} "
                            + f"vs {input_dict[name]['material']})!")
            if isinstance(ee, Geant4CollimatorTip) or 'tip_material' in input_dict[name]:
                if not isinstance(ee, Geant4CollimatorTip):
                    self.stop()
                    raise ValueError(f"Element {name} is not a Geant4CollimatorTip "
                                    + "in the line, but it has tip material in the input file!")
                if 'tip_material' not in input_dict[name] or 'tip_thickness' not in input_dict[name]:
                    self.stop()
                    raise ValueError(f"Element {name} is a Geant4CollimatorTip, "
                                    + "but it has no tip material in the input file!")
                if ee.tip_material.geant4_name != input_dict[name]['tip_material']:
                    self._print(f"Warning: Tip material of {name} differs from input file "
                            + f"({ee.tip_material.geant4_name} vs {input_dict[name]['tip_material']})! Overwritten.")
                    ee.tip_material.geant4_name = input_dict[name]['tip_material']
                if not np.isclose(ee.length, input_dict[name]['tip_thickness'], atol=1e-9):
                    self._print(f"Warning: Tip thickness of {name} differs from input file "
                            + f"({ee.tip_thickness} vs {input_dict[name]['tip_thickness']})! Overwritten.")
                    ee.tip_thickness = input_dict[name]['tip_thickness']
            jaw = input_dict[name]['jaw']
            if jaw is not None and not hasattr(jaw, '__iter__'):
                jaw = [jaw, -jaw]
            if jaw is None or (jaw[0] is None and jaw[1] is None):
                ee.jaw = None
            else:
                if jaw[0] is None:
                    if ee.side != 'right':
                        self._print(f"Warning: {name} is right-sided in the input file, "
                                + "but not in the line! Overwritten by the former.")
                        ee.side = 'right'
                elif ee.jaw_L is None or not np.isclose(ee.jaw_L, jaw[0], atol=1e-9):
                    self._print(f"Warning: Jaw_L of {name} differs from input file "
                            + f"({ee.jaw_L} vs {jaw[0]})! Overwritten.")
                    ee.jaw_L = jaw[0]
                if jaw[1] is None:
                    if ee.side != 'left':
                        self._print(f"Warning: {name} is left-sided in the input file, "
                                + f"but not in the line! Overwritten by the former.")
                        ee.side = 'left'
                elif ee.jaw_R is None or not np.isclose(ee.jaw_R, jaw[1], atol=1e-9):
                    self._print(f"Warning: Jaw_R of {name} differs from input file "
                            + f"({ee.jaw_R} vs {jaw[1]})! Overwritten.")
                    ee.jaw_R = jaw[1]
            tilt = input_dict[name]['tilt']
            if not hasattr(tilt, '__iter__'):
                tilt = [tilt, -tilt]
            if ee.side != 'right' and not np.isclose(ee.tilt_L, tilt[0], atol=1e-9):
                self._print(f"Warning: Tilt_L of {name} differs from input file "
                        + f"({ee.tilt_L} vs {tilt[0]})! Overwritten.")
                ee.tilt_L = tilt[0]
            if ee.side != 'left' and not np.isclose(ee.tilt_R, tilt[1], atol=1e-9):
                self._print(f"Warning: Tilt_R of {name} differs from input file "
                        + f"({ee.tilt_R} vs {tilt[1]})! Overwritten.")
                ee.tilt_R = tilt[1]



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
from warnings import warn

import xtrack as xt
import xobjects as xo
import xpart as xp

from .beam_elements import (collimator_classes, BaseCrystal, Geant4Collimator,
                            Geant4Crystal, FlukaCollimator, FlukaCrystal)


def generate_pencil_on_collimator(line, name, num_particles, *, side='+-', pencil_spread=1e-6,
                                  impact_parameter=0, sigma_z=7.61e-2, twiss=None, longitudinal=None,
                                  longitudinal_betatron_cut=None, _capacity=None, tw=None,
                                  _longitudinal_coords=None, **kwargs):
    """Generate a pencil beam on a collimator."""

    # Do some general checks
    if not line._has_valid_tracker():
        raise ValueError("Please build tracker before generating pencil distribution!")
    coll = line[name]
    if not isinstance(coll, tuple(collimator_classes)):
        raise ValueError("Need to provide a valid collimator!")
    if coll.optics is None:
        raise ValueError("Need to assign optics to collimators before generating pencil distribution!")

    num_particles = int(num_particles)
    if _capacity is None and len(line.get_elements_of_type(
                                (FlukaCollimator, FlukaCrystal))[0]) > 0:
        import xcoll as xc
        _capacity = cap if (cap := xc.fluka.engine.capacity) else 5*num_particles
    if _capacity is None and len(line.get_elements_of_type(
                                (Geant4Collimator, Geant4Crystal))[0]) > 0:
        _capacity = 5*num_particles

    # Define the plane
    angle = coll.angle
    if abs(np.mod(angle-90,180)-90) < 1e-6:
        plane = 'x'
        transv_plane = 'y'
    elif abs(np.mod(angle,180)-90) < 1e-6:
        plane = 'y'
        transv_plane = 'x'
    else:
        raise NotImplementedError("Pencil beam on a skew collimator not yet supported!")

    if coll.side == 'left':
        if side == '-':
            raise ValueError("Cannot generate a pencil on the right jaw of a left-"
                             "sided collimator!")
        side = '+'
    if coll.side == 'right':
        if side == '+':
            raise ValueError("Cannot generate a pencil on the left jaw of a right-"
                             "sided collimator!")
        side = '-'

    if tw is not None:
        warn("The argument tw is deprecated. Please use twiss instead.", FutureWarning)
        if twiss is None:
            twiss = tw

    if twiss is None:
        twiss = line.twiss()

    # Longitudinal plane
    if _longitudinal_coords:
        zeta = _longitudinal_coords[0]
        delta = _longitudinal_coords[1]

    else:
        zeta, delta = _generate_longitudinal_dist(line, num_particles, sigma_z, longitudinal)

        # Generate 4D coordinates
        # TODO: there is some looping in the calculation here and in xpart. Can it be improved?
        if side == '+-':
            num_plus = int(num_particles/2)
            num_min  = int(num_particles - num_plus)
            zeta_plus = zeta[:num_plus] if hasattr(zeta, '__iter__') else zeta
            zeta_min  = zeta[num_plus:] if hasattr(zeta, '__iter__') else zeta
            delta_plus = delta[:num_plus] if hasattr(delta, '__iter__') else delta
            delta_min  = delta[num_plus:] if hasattr(delta, '__iter__') else delta
            if _capacity:
                _capacity_plus = int(_capacity/2)
                _capacity_min  = int(_capacity - _capacity_plus)
            else:
                _capacity_plus = None
                _capacity_min  = None
            part_plus = generate_pencil_on_collimator(line=line, name=name, num_particles=num_plus,
                                impact_parameter=impact_parameter, _capacity=_capacity_plus,
                                side='+', pencil_spread=pencil_spread, twiss=twiss,
                                _longitudinal_coords=[zeta_plus, delta_plus],
                                **kwargs)
            part_min = generate_pencil_on_collimator(line=line, name=name, num_particles=num_min,
                                impact_parameter=impact_parameter, _capacity=_capacity_min,
                                side='-', pencil_spread=pencil_spread, twiss=twiss,
                                _longitudinal_coords=[zeta_min, delta_min],
                                **kwargs)

            part = xt.Particles.merge([part_plus, part_min])
            part.start_tracking_at_element = part_plus.start_tracking_at_element
            assert part.start_tracking_at_element == part_min.start_tracking_at_element
            return part

    pencil, p_pencil, transverse_norm, p_transverse_norm, is_converging, at_element = \
                    _generate_4D_pencil_one_jaw(line, name, num_particles, plane, side,
                                                impact_parameter, pencil_spread, twiss, **kwargs)

    # Build the particles
    if plane == 'x':
        part = line.build_particles(
                x=pencil, px=p_pencil, y_norm=transverse_norm, py_norm=p_transverse_norm,
                zeta=zeta, delta=delta, nemitt_x=coll.nemitt_x, nemitt_y=coll.nemitt_y,
                at_element=at_element, _capacity=_capacity, _context=coll._buffer.context,
                **kwargs)
    else:
        part = line.build_particles(
                x_norm=transverse_norm, px_norm=p_transverse_norm, y=pencil, py=p_pencil,
                zeta=zeta, delta=delta, nemitt_x=coll.nemitt_x, nemitt_y=coll.nemitt_y,
                at_element=at_element, _capacity=_capacity, _context=coll._buffer.context,
                **kwargs)

    part._init_random_number_generator()

    if not is_converging:
        dri = xt.Drift(length=-coll.length)
        dri.track(part)
        part.start_tracking_at_element -= 1
        part.at_element -= 1

    return part


def generate_delta_from_dispersion(line, at_element, *, plane, position_mm, nemitt_x, nemitt_y,
                                   twiss=None, betatron_cut=0, match_at_front=True):
    if line.tracker is None:
        raise ValueError("Need to build tracker first!")
    if not hasattr(betatron_cut, '__iter__'):
        if hasattr(position_mm, '__iter__'):
            betatron_cut = np.full_like(position_mm, betatron_cut)
    elif not hasattr(position_mm, '__iter__'):
        position_mm = np.full_like(betatron_cut, position_mm)
    elif len(position_mm) != len(betatron_cut):
        raise ValueError
    if plane not in ['x', 'y']:
        raise ValueError("The variable 'plane' needs to be either 'x' or 'y'!")

    if twiss is None:
        twiss = line.twiss()

    beam_sizes = twiss.get_beam_covariance(nemitt_x=nemitt_x, nemitt_y=nemitt_y)
    beam_sizes = beam_sizes.rows[at_element:f'{at_element}>>1'][f'sigma_{plane}']
    sigma = beam_sizes[0] if match_at_front else beam_sizes[1]
    delta  = (position_mm - betatron_cut*sigma - twiss.rows[at_element][plane])
    delta /= twiss.rows[at_element][f'd{plane}']

    return delta


def _generate_4D_pencil_one_jaw(line, name, num_particles, plane, side, impact_parameter,
                                pencil_spread, twiss=None, **kwargs):
    coll = line[name]
    beam_sizes = twiss.get_beam_covariance(nemitt_x=coll.nemitt_x, nemitt_y=coll.nemitt_y)

    # Is it converging or diverging?
    # TODO: dispersion might change this
    # TODO: skew collimators
    tolerance_tilt = 1e-12 # 0.1 urad tolerance on jaw tilt  =>  we prioritise converging
    divergence = coll.divergence
    if side == '+':
        if isinstance(coll, BaseCrystal):
            # A pencil on the crystal should always be upstream
            is_converging = True
            pencil_pos = coll.jaw_U + impact_parameter
        else:
            betatron_angle = coll.gap_L * divergence
            is_converging = coll.tilt_L + tolerance_tilt >= betatron_angle
            print(f"Left jaw of collimator {name} is {'con' if is_converging else 'di'}verging.")
            if is_converging:
                pencil_pos = coll.jaw_LU + impact_parameter
            else:
                pencil_pos = coll.jaw_LD + impact_parameter
    elif side == '-':
        if isinstance(coll, BaseCrystal):
            # A pencil on the crystal should always be upstream
            is_converging = True
            pencil_pos = coll.jaw_U - impact_parameter
        else:
            betatron_angle = coll.gap_R * divergence
            is_converging = coll.tilt_R - tolerance_tilt <= betatron_angle
            print(f"Right jaw of collimator {name} is {'con' if is_converging else 'di'}verging.")
            if is_converging:
                pencil_pos = coll.jaw_RU - impact_parameter
            else:
                pencil_pos = coll.jaw_RD - impact_parameter
    else:
        raise ValueError(f"Sinde {side} not supported in  _generate_4D_pencil_one_jaw!")

    if is_converging:
        # pencil at front of jaw
        sigma = beam_sizes.rows[name:f'{name}>>1'][f'sigma_{plane}'][0]
        tw_at_s = twiss.rows[name]
        at_element = name
    else:
        # pencil at back of jaw
        sigma = beam_sizes.rows[name:f'{name}>>1'][f'sigma_{plane}'][1]
        tw_at_s = twiss.rows[f'{name}>>1']
        at_element = line.element_names[line.element_names.index(name)+1]

    dr_sigmas = pencil_spread/sigma

    # Collimator plane: generate pencil distribution
    pencil, p_pencil = xp.generate_2D_pencil_with_absolute_cut(
                        num_particles, plane=plane, absolute_cut=pencil_pos, line=line,
                        dr_sigmas=dr_sigmas, nemitt_x=coll.nemitt_x, nemitt_y=coll.nemitt_y,
                        at_element=at_element, side=side, twiss=tw_at_s, **kwargs
                       )

    # Other plane: generate gaussian distribution in normalized coordinates
    transverse_norm   = np.random.normal(size=num_particles)
    p_transverse_norm = np.random.normal(size=num_particles)

    return pencil, p_pencil, transverse_norm, p_transverse_norm, is_converging, at_element


def _generate_longitudinal_dist(line, num_particles, sigma_z, longitudinal):
    # TODO: make this more general, make this better
    if longitudinal is None:
        return 0, 0
    elif longitudinal == 'matched_dispersion':
        raise NotImplementedError
        # if longitudinal_betatron_cut is None:
        #     cut = 0
        # else:
        #     cut = np.random.uniform(-longitudinal_betatron_cut, longitudinal_betatron_cut,
        #                             num_particles)
        # delta = generate_delta_from_dispersion(line, name, plane=plane, position_mm=pencil,
        #                                        nemitt_x=nemitt_x, nemitt_y=nemitt_y, twiss=tw,
        #                                        betatron_cut=cut, match_at_front=is_converging)
        # zeta  = 0
    elif longitudinal == 'bucket':
        return xp.generate_longitudinal_coordinates(
                num_particles=num_particles, distribution='gaussian', sigma_z=sigma_z, line=line
        )
    elif not hasattr(longitudinal, '__iter__'):
        raise ValueError
    elif len(longitudinal) != 2:
        raise ValueError
    elif isinstance(longitudinal, str):
        raise ValueError
    elif isinstance(longitudinal, dict):
        return longitudinal['zeta'], longitudinal['delta']
    else:
        return longitudinal[0], longitudinal[1]



import os

def create_FEDB():
    """
    definition of fedb structure
    """
    # global path ( i.e. to present directory):
    global_path = os.path.realpath(os.path.dirname( __file__ ))

    # ...in order to import FEDB class:
    from SUPPORT import FEDB

    # instance of FEDB:
    present_FEDB = FEDB()

    # set global path:
    present_FEDB.set_global_path( global_path )

    # - naming convention:
    present_FEDB.set_naming_convention( "%s_%s" )

    # - materials file:
    present_FEDB.set_material_file_subpath( "materials/materials.inp" )

    # - SUBPATHs:
    present_FEDB.set_SUBPATHs( "body", "bodies/"     + present_FEDB.ret_naming_convention() + ".bodies" )
    present_FEDB.set_SUBPATHs( "regn", "regions/"    + present_FEDB.ret_naming_convention() + ".regions" )
    present_FEDB.set_SUBPATHs( "assm", "materials/"  + present_FEDB.ret_naming_convention() + ".assignmat" )
    present_FEDB.set_SUBPATHs( "stpz", "stepsizes/"  + present_FEDB.ret_naming_convention() + ".stepsizes" )
    present_FEDB.set_SUBPATHs( "smbl", "assemblies/" + present_FEDB.ret_naming_convention() + ".lbp" )

    # - inflect material files:
    present_FEDB.inflect_materials()

    # check integrity of declared structure:
    present_FEDB.check()

    return present_FEDB

# ------------------------------------------------------------------------------
present_FEDB = create_FEDB()
# ------------------------------------------------------------------------------



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import sys
from math import floor, log10
from subprocess import run, PIPE

try:
    from xaux import FsPath  # TODO: once xaux is in Xsuite keep only this
except (ImportError, ModuleNotFoundError):
    from ...xaux import FsPath

from ..environment import BaseEnvironment
from ...general import _pkg_root


_FORTRAN_SRC   = FsPath(_pkg_root / 'scattering_routines' / 'fluka' / 'FORTRAN_src').resolve()
_FEDB_TEMPLATE = FsPath(_pkg_root / 'scattering_routines' / 'fluka' / 'fedb').resolve()


class FlukaEnvironment(BaseEnvironment):
    # The paths to be set. The value is the parent depth that needs to be brute-forced (0 = file itself, None = no brute-force)
    _paths = {'fluka': 2, 'flukaserver': 1, 'linebuilder': 0}
    _optional_paths = {'flair': 1}
    _read_only_paths = {'fedb': 0}

    def __init__(self):
        super().__init__()
        # Initialize the FEDB
        self._init_fedb()

    @property
    def fedb(self):
        return self._data_dir / 'fedb'

    @property
    def compiled(self):
        try:
            from pyflukaf import track_fluka
            return True
        except (ModuleNotFoundError, ImportError) as error:
            return False

    def compile(self, flukaio_lib=None, flukaio_path=None, verbose=False):
        # Check all dependencies
        self.assert_installed('make', verbose=verbose)
        self.assert_installed('cmake', verbose=verbose)
        self.assert_gcc_installed(verbose=verbose)
        self.assert_gfortran_installed(verbose=verbose)
        cwd = FsPath.cwd()
        self.store_environment()

        # Get FlukaIO
        if flukaio_lib is None:
            # Check the provided FlukaIO path
            if flukaio_path is None:
                raise ValueError("FlukaIO path must be provided!")
            flukaio_path = FsPath(flukaio_path).resolve()
            if not flukaio_path.exists():
                raise FileNotFoundError(f"Could not find FlukaIO path {flukaio_path}!")
            self.brute_force_path(flukaio_path)
            # Copy the FlukaIO files to a temporary directory and compile it
            dest = (self.temp_dir / 'flukaio').resolve()
            dest.mkdir(parents=True, exist_ok=True)
            for path in flukaio_path.glob('*'):
                if path.name == '.git' or path.name == 'docs' or path.name == 'tests' \
                or path.name == 'samples':
                    continue
                else:
                    path.copy_to(dest, method='mount')
            os.chdir(dest)
            cmd = run(['make', 'clean'], stdout=PIPE, stderr=PIPE)
            cmd = run(['make', 'libs', 'BUILD64=Y'], stdout=PIPE, stderr=PIPE)
            if cmd.returncode == 0:
                if verbose:
                    print("Compiled FlukaIO successfully.")
            else:
                stderr = cmd.stderr.decode('UTF-8').strip()
                os.chdir(cwd)
                raise RuntimeError(f"Failed to compile FlukaIO!\nError given is:\n{stderr}")
            flukaio_lib = dest  / 'lib' / 'libFlukaIO64.a'
        flukaio_lib = FsPath(flukaio_lib).resolve()
        if not flukaio_lib.exists():
            raise FileNotFoundError(f"Could not find FlukaIO library {flukaio_lib}!")

        # Copy the FORTRAN source files to the temporary directory and compile it
        dest = (self.temp_dir / 'FORTRAN_src').resolve()
        dest.mkdir(parents=True, exist_ok=True)
        flukaio_lib.copy_to(dest / 'libFlukaIO.a', method='mount')
        for path in _FORTRAN_SRC.glob('*'):
            path.copy_to(dest, method='mount')
        os.chdir(dest)
        cmd = run(['python', '-m', 'numpy.f2py', '-m', 'pyflukaf', 'pyfluka.f90',
                   '--quiet'], stdout=PIPE, stderr=PIPE)
        if cmd.returncode == 0:
            if verbose:
                print("Created numpy FORTRAN headers.")
        else:
            stderr = cmd.stderr.decode('UTF-8').strip()
            os.chdir(cwd)
            raise RuntimeError(f"Failed to create numpy FORTRAN headers!\nError given is:\n{stderr}")
        cmd = run(['meson', 'setup', 'build'], stdout=PIPE, stderr=PIPE)
        if cmd.returncode == 0:
            if verbose:
                print("Setup meson build successfully.")
        else:
            stdout = cmd.stdout.decode('UTF-8').strip()
            stderr = cmd.stderr.decode('UTF-8').strip()
            os.chdir(cwd)
            raise RuntimeError(f"Failed to setup meson build!\n"
                               f"Output given is:\n{stdout}"
                               f"Error given is:\n{stderr}")
        cmd = run(["meson", "compile", "-C", "build"], stdout=PIPE, stderr=PIPE)
        if cmd.returncode == 0:
            if verbose:
                print(cmd.stdout.decode('UTF-8').strip())
                print("Compiled pyflukaf successfully.")
        else:
            stdout = cmd.stdout.decode('UTF-8').strip()
            stderr = cmd.stderr.decode('UTF-8').strip()
            os.chdir(cwd)
            raise RuntimeError(f"Failed to compile pyflukaf!\n"
                               f"Output given is:\n{stdout}"
                               f"Error given is:\n{stderr}")
        os.chdir(cwd)
        # Collect the compiled shared library
        so = list((self.temp_dir / 'FORTRAN_src' / 'build').glob('pyflukaf.*so'))
        if len(so) > 1:
            raise RuntimeError(f"Compiled into multiple pyflukaf shared libraries!")
        if len(so) == 0:
            raise RuntimeError(f"Failed pyFLUKA compilation! No shared library found in "
                             + f"{self.data_dir.as_posix()}!")
        so = so[0]
        so.move_to(self.data_dir / so.name)
        if verbose:
            print(f"Created pyFLUKA shared library in {so}.")
        # Clean up the temporary directory
        self.temp_dir = None
        self.restore_environment()


    def import_fedb(self, fedb_path, verbose=False, overwrite=False):
        import xcoll as xc
        fedb_path = FsPath(fedb_path)
        if not fedb_path.exists():
            raise FileNotFoundError(f"Could not find FEDB {fedb_path}!")
        self._init_fedb(overwrite=True)  # Always overwrite the build-in FEDB (for in-between modifications in Xcoll)
        # Get assemblies
        for file in fedb_path.glob(f'assemblies/*.lbp'):
            if (self.fedb / 'assemblies' / file.name).exists() and not overwrite:
                if verbose:
                    print(f"Warning: Assembly {file.name} already exists in Xcoll. "
                         + "Use overwrite=True to overwrite it.")
                continue
            meta = _FEDB_TEMPLATE / 'metadata' / 'coupling' / f'{file.name}.json'
            if meta.exists():
                pro = xc.FlukaAssembly.from_json(meta)
                if verbose:
                    print(f"Imported assembly {pro.fedb_series}_{pro.fedb_tag}")
            else:
                if verbose:
                    print(f"Warning: No metadata found for assembly {file.name} in {fedb_path}!")
                part = file.stem.split('_')
                fedb_series = part[0]
                fedb_tag = '_'.join(part[1:])
                pro = xc.FlukaAssembly(fedb_series, fedb_tag)
            pro.assembly_file = file
        # Get prototypes
        for file in fedb_path.glob(f'bodies/*.bodies'):
            if (self.fedb / 'bodies' / file.name).exists() and not overwrite:
                if verbose:
                    print(f"Warning: Prototype {file.name} already exists in Xcoll. "
                         + "Use overwrite=True to overwrite it.")
                continue
            meta = _FEDB_TEMPLATE / 'metadata' / 'coupling' / f'{file.name}.json'
            if meta.exists():
                pro = xc.FlukaPrototype.from_json(meta)
                if verbose:
                    print(f"Imported prototype {pro.fedb_series}_{pro.fedb_tag}")
            else:
                if verbose:
                    print(f"Warning: No metadata found for prototype {file.name} in {fedb_path}!")
                part = file.stem.split('_')
                fedb_series = part[0]
                fedb_tag = '_'.join(part[1:])
                pro = xc.FlukaPrototype(fedb_series, fedb_tag)
            pro.body_file = file
            pro.material_file = fedb_path / 'materials' / f'{file.stem}.assignmat'
            pro.region_file = fedb_path / 'regions' / f'{file.stem}.regions'
        # Get stepsizes (currently not used)
        for file in fedb_path.glob(f'stepsizes/*'):
            file.copy_to(self.fedb / 'stepsizes' / file.name, method='mount')
        # Link the new files into the registry
        self._init_fedb()


    def set_fluka_environment(self):
        self.brute_force_path('fluka')
        self.brute_force_path('flukaserver')

    def set_flair_environment(self):
        self.brute_force_path('flair')

    def set_fedb_environment(self, fedb=None):
        self.store_environment()
        if fedb is None:
            fedb = self.fedb
        self.brute_force_path(fedb)
        self.brute_force_path('linebuilder')
        os.environ['FEDB_PATH'] = fedb.as_posix()
        os.environ['LB_PATH'] = self.linebuilder.as_posix()
        # Brute-force the system paths
        sys.path.insert(0, (self.linebuilder / "src").as_posix())
        sys.path.insert(0, (self.linebuilder / "lib").as_posix())
        sys.path.insert(0, fedb.as_posix())


    def create_temp_fedb(self, assemblies):
        fedb = FsPath('temp_fedb').resolve()
        fedb.mkdir(parents=True)
        (fedb / 'assemblies').mkdir(parents=True)
        (fedb / 'bodies').mkdir(parents=True)
        (fedb / 'regions').mkdir(parents=True)
        (fedb / 'materials').mkdir(parents=True)
        (fedb / 'stepsizes').mkdir(parents=True)
        (fedb / 'tools').symlink_to(self.fedb / 'tools')
        (self.fedb / 'structure.py').copy_to(fedb / 'structure.py')
        (self.fedb / 'materials' / 'materials.inp').copy_to(fedb / 'materials' / 'materials.inp')
        for assm in assemblies:
            assm.populate_into_temp_fedb(fedb)
        return fedb


    def run_flair(self, input_file=None):
        if input_file is None:
            return
        input_file = FsPath(input_file)
        try:
            self.set_flair_environment()
        except FileNotFoundError:
            print("Flair not found. Cannot view input file.")
            return
        cmd = run([self.flair.as_posix(), input_file.as_posix()],
                  stdout=PIPE, stderr=PIPE)
        if cmd.returncode != 0:
            stderr = cmd.stderr.decode('UTF-8').strip()
            raise RuntimeError(f"Failed to run flair on input file {input_file}!\n"
                             + f"Error given is:\n{stderr}")


    def test_assembly(self, fedb_series, fedb_tag, *, show=True, keep_files=False):
        if show:
            try:
                self.set_flair_environment()
            except FileNotFoundError:
                print("Flair not found. Cannot view assembly.")
                return
        else:
            keep_files=True
        self.set_fedb_environment()
        cmd = run(['python', self.fedb / 'tools' / 'test_assembly.py', fedb_series,
                   fedb_tag], stdout=PIPE, stderr=PIPE)
        self.restore_environment()
        if cmd.returncode != 0:
            stderr = cmd.stderr.decode('UTF-8').strip()
            raise RuntimeError(f"Failed to run flair!\nError given is:\n{stderr}")
        file = FsPath.cwd() / f"{fedb_series}_{fedb_tag}.inp"
        if not file.exists():
            raise FileNotFoundError(f"Temporary input file {file} not generated!")
        logfile = file.with_suffix('.log')
        if not logfile.exists():
            raise FileNotFoundError(f"Temporary log file {logfile} not generated!")
        self.run_flair(file)
        if not keep_files:
            file.unlink()
            logfile.unlink()


    # =======================
    # === Private Methods ===
    # =======================

    def _init_fedb(self, overwrite=False):
        from xcoll import FlukaPrototype
        self.fedb.mkdir(parents=True, exist_ok=True)
        for directory in ['assemblies', 'bodies', 'regions', 'materials', 'stepsizes', 'metadata']:
            (self.fedb / directory).mkdir(parents=True, exist_ok=True)
            for f in (_FEDB_TEMPLATE / directory).glob('*.*'):
                new_file = self.fedb / directory / f.name
                if new_file.exists() and overwrite:
                    new_file.unlink()
                if not new_file.exists():
                    f.copy_to(new_file, method='mount')
        tools = self.fedb / 'tools'
        if tools.exists() and overwrite:
            tools.rmtree()
        if not tools.exists():
            (_FEDB_TEMPLATE / 'tools').copy_to(tools.parent, method='mount')
        structure = self.fedb / 'structure.py'
        if structure.exists() and overwrite:
            structure.unlink()
        if not structure.exists():
            (_FEDB_TEMPLATE / 'structure.py').copy_to(structure, method='mount')
        prototypes = (self.fedb / 'metadata').glob('*_*.json')
        for file in prototypes:
            FlukaPrototype.from_json(file)


def format_fluka_float(value):
    if abs(value) > 9.99e99:
        raise ValueError(f"Value {value} is too large for FLUKA.")
    elif abs(value) < 1e-99:
        return f"{'0.0':>10}"
    if 1.e-4 <= value <= 999999999. or -1.e-4 >= value >= -99999999.:
        max_digits = 8 if value > 0 else 7
        n_decimal_digits = int(max_digits - max(floor(log10(abs(value))), 0))
        value_string = f"{value:10.{n_decimal_digits}f}"
        value_splitted = value_string.split('.')
        if len(value_splitted) == 1:
            return f"{value_string[1:]}."
        else:
            decimals = value_splitted[1]
            while decimals[-1] == '0':
                decimals = decimals[:-1]
                if decimals == '':
                    decimals = '0'
                    break
            value_string = f"{value_splitted[0]}.{decimals}"
            return f"{value_string:>10}"
    else:
        max_digits = 4 if value > 0 else 3
        value_string = f"{value:.{max_digits}E}"
        value_splitted = value_string.split('E')
        value_splitted_mant = value_splitted[0].split('.')
        if len(value_splitted_mant) == 1:
            return value_string
        else:
            decimals = value_splitted_mant[1]
            dot = '.'
            while decimals[-1] == '0':
                decimals = decimals[:-1]
                if decimals == '':
                    dot = ''
                    break
            value_string = f"{value_splitted_mant[0]}{dot}{decimals}E{value_splitted[1]}"
            return f"{value_string:>10}"



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .engine import Geant4Engine
from .track import track_pre, track_core, track_post



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import io
import numpy as np

from xtrack import Particles

from ...constants import (LOST_WITHOUT_SPEC, LOST_ON_GEANT4_COLL,
    EXCITED_ION_STATE, MASSLESS_OR_NEUTRAL, VIRTUAL_ENERGY, HIT_ON_GEANT4_COLL)


def track_pre(coll, particles):
    import xcoll as xc

    # Initialize ionisation loss accumulation variable
    if coll._acc_ionisation_loss < 0:
        coll._acc_ionisation_loss = 0.

    if not xc.geant4.engine.assert_ready_to_track_or_skip(
    coll, particles, _necessary_attributes=['geant4_id']):
        return False  # Stop tracking

    return True  # Continue tracking


def track_post(coll, particles):
    alive_states = np.unique(particles.state[particles.state > 0])
    assert len(alive_states) <= 1, f"Unexpected alive particle states after tracking: {alive_states}"
    if len(alive_states) == 1:
        assert alive_states[0] == 1, f"Unexpected alive particle state after tracking: {alive_states[0]}"


# TODO: need to rework this logic with HIT_ON_GEANT4_COLL


def track_core(coll, part):
    import xcoll as xc
    xc.geant4.engine._g4link.clearData() # Clear the old data - bunch particles and hits

    # send_to_geant4 = part.state == HIT_ON_GEANT4_COLL
    send_to_geant4 = part.state > 0
    # npart          = send_to_geant4.sum()
    # max_id         = part.particle_id[part.state > -9999].max()
    # assert npart  <= part._num_active_particles
    # if npart == 0:
    #     return

    num_sent = send_to_geant4.sum()

    coll._drift(part, -250e-9) # Margin before Geant4 collimator; added by BDSIM geometry

    # Get particle data
    q0 = part.q0
    m0 = part.mass0
    p0c = part.p0c[0]
    beta0 = part.beta0[0]
    s_in = part.s[send_to_geant4][0]
    ele_in = part.at_element[send_to_geant4][0]
    turn_in = part.at_turn[send_to_geant4][0]
    precision  = p0c * 1.e-12  # To avoid numerical issues like negative enervy. Ideally this should be 2.22e-15

    rpp  = part.rpp[send_to_geant4]
    x    = part.x[send_to_geant4]
    xp   = part.px[send_to_geant4] * rpp   # Director cosine
    y    = part.y[send_to_geant4]
    yp   = part.py[send_to_geant4] * rpp   # Director cosine
    zeta = part.zeta[send_to_geant4]
    p    = p0c * (1 + part.delta[send_to_geant4]) * part.mass_ratio[send_to_geant4]
    q    = q0 * part.charge_ratio[send_to_geant4]
    weight = part.weight[send_to_geant4]
    pdgid  = part.pdg_id[send_to_geant4]
    pid    = part.particle_id[send_to_geant4]

    if xc.geant4.engine.reentry_protection_enabled:
        ### remove this part after geant4 bug fixed
        coords = {'x': x, 'xp': xp, 'y': y, 'yp': yp, 'zeta': zeta, 'p': p,
                  'q': q, 'weight': weight, 'pdgid': pdgid, 'id': pid}
        buf = io.BytesIO() # Use numpy.savez to serialize
        np.savez(buf, **coords)
        buf.seek(0)
        result_blob = xc.geant4.engine._g4link.add_particles_and_collimate_return(
                            buf.getvalue(), coll.geant4_id, num_sent)
        result_buf = io.BytesIO(result_blob) # Deserialize
        products = np.load(result_buf)
    else:
        coords = [x, xp, y, yp, zeta, p, q, weight, pdgid, pid]
        xc.geant4.engine._g4link.addParticles(coords)
        xc.geant4.engine._g4link.selectCollimator(coll.geant4_id)
        xc.geant4.engine._g4link.collimate()
        products = xc.geant4.engine._g4link.collimateReturn(num_sent)

    # Careful with all the masking!
    # Double-mask assignment does not work, e.g. part.state[mask1][mask2] = 1 will do nothing...

    # Kill particles that died just now
    returned_dead = products['state'][:num_sent] == LOST_WITHOUT_SPEC
    idx_dead = np.arange(len(part.x))[send_to_geant4][returned_dead]
    part.state[idx_dead] = -LOST_ON_GEANT4_COLL

    # Update particles that are still alive
    returned_alive = products['state'][:num_sent] > 0
    idx_alive = np.arange(len(part.x))[send_to_geant4][returned_alive]    # BDSIM keeps particle order
    idx_returned_alive = np.nonzero(returned_alive)[0]

    # Energy needs special treatment
    m_in = part.mass[idx_alive]
    new_p = products['p'][idx_returned_alive]
    new_energy = np.sqrt(new_p**2 + m_in**2)
    E_diff = np.zeros(len(part.x))
    E_diff[idx_alive] = part.energy[idx_alive] - new_energy
    if np.any(E_diff < -precision):
        raise ValueError(f"Geant4 returned particle with energy higher than incoming particle!")
    E_diff[E_diff < precision] = 0. # Lower cut on energy loss
    part.add_to_energy(-E_diff)
    part.weight[idx_alive] = products['weight'][idx_returned_alive]
    coll._acc_ionisation_loss += np.sum(E_diff[idx_alive]*part.weight[idx_alive])

    rpp  = part.rpp[idx_alive]
    part.x[idx_alive]      = products['x'][idx_returned_alive]
    part.px[idx_alive]     = products['xp'][idx_returned_alive] / rpp   # Director cosine back to px
    part.y[idx_alive]      = products['y'][idx_returned_alive]
    part.py[idx_alive]     = products['yp'][idx_returned_alive] / rpp   # Director cosine back to py
    part.zeta[idx_alive]   = products['zeta'][idx_returned_alive]

    # Add new particles created in Geant4
    q_new = products['q'][num_sent:]
    pdg_id = products['pdg_id'][num_sent:]
    mask_new = xc.geant4.engine._mask_particle_return_types(pdg_id, q_new)
    idx_new = np.nonzero(mask_new)[0] + num_sent
    assert np.all(products['state'][idx_new] == 1)

    if np.any(mask_new):
        # Check that there is enough room in the particles object
        num_assigned = part._num_lost_particles + part._num_active_particles
        num_free = part._capacity - num_assigned
        num_needed = mask_new.sum()
        if num_free < num_needed:
            raise RuntimeError(f"Too many particles generated by Geant4 ({num_needed} needed, "
                             + f"but only {num_free} free in particles object)!")

        # Parent particle IDs
        parents = products['parent_particle_id'][idx_new]
        pids = part.particle_id[send_to_geant4]
        max_id_in = int(pids.max())
        assert np.all(parents >= 0)        # Sanity check
        assert parents.max() <= max_id_in  # Sanity check

        # Build dense lookup table: ID -> index in `pids`
        lookup = np.full(max_id_in + 1, -1, dtype=np.int64)
        lookup[pids] = np.arange(pids.size, dtype=np.int64)
        idx_parents = lookup[parents]   # MUCH faster than solution with np.where
        if np.any(idx_parents < 0):
            missing = np.unique(parents[idx_parents < 0])
            raise RuntimeError(f"Parent IDs not found in particle_id: {missing}")

        # Mass
        m_new = products['m'][idx_new]
        if np.any(m_new < -precision):
            raise ValueError(f"Geant4 returned particle with negative mass!")
        massless = np.abs(m_new) < 1.e-12
        m_new[massless] = m0  # TODO

        # Charge
        q_new = products['q'][idx_new]
        neutral = np.abs(q_new) < 1.e-12
        q_new[massless | neutral] = q0  # TODO

        # Energy
        p_new = products['p'][idx_new]
        delta = np.zeros_like(p_new)
        delta[massless] = p_new[massless]/p0c - 1
        delta[~massless] = p_new[~massless]/p0c * m0/m_new[~massless] - 1

        new_part = Particles(_context=part._buffer.context,
                p0c = part.p0c[0],
                mass0 = part.mass0,
                q0 = part.q0,
                s = s_in + coll.length,
                x = products['x'][idx_new],
                px = products['xp'][idx_new] * (1 + delta), # Director cosine back to px
                y = products['y'][idx_new],
                py = products['yp'][idx_new] * (1 + delta), # Director cosine back to py
                zeta = products['zeta'][idx_new],
                delta = delta,
                mass_ratio = m_new/m0,
                charge_ratio = q_new/q0,
                at_element = ele_in,
                at_turn = turn_in,
                parent_particle_id = products['parent_particle_id'][idx_new],
                pdg_id = products['pdg_id'][idx_new],
                weight = products['weight'][idx_new]
        )

        # Correct the deposited energy of parent particles: not everything was lost there.
        E_children = np.bincount(idx_parents, weights=new_part.energy, minlength=part._capacity)
        if np.any(E_children < -precision):
            raise ValueError(f"Geant4 returned particles with summed energy higher than parent particle!")
        # If the parent survived, this should not be done but the energy should be subtracted
        # from the accumulated ionisation loss (as it is accounted for by the child)
        mask_parent_survived = (part.state==1) & (E_children > 0)
        if np.any(mask_parent_survived):
            if np.any(E_diff[mask_parent_survived] - E_children[mask_parent_survived]) < -precision:
                raise ValueError(f"Geant4 returned children with a surviving parent, however, there "
                               + f"was a larger energy loss than the children energy!")
            coll._acc_ionisation_loss -= np.sum(E_children[mask_parent_survived]*part.weight[mask_parent_survived])
            E_children[mask_parent_survived] = 0.
        # If the deposited energy is lower than the rest mass, it cannot be represented by the original
        # particle. We make a virtual particle with fake mass (E=2m) to avoid negative square roots.
        mask_virtual = (part.energy - E_children < part.mass) & (E_children > 0)
        if np.any(mask_virtual):
            # import ipdb; ipdb.set_trace()
            virtual_mass = (part.energy - E_children)[mask_virtual] / np.sqrt(2)
            new_ptau = part.ptau.copy()
            old_mass = part.mass[mask_virtual]
            old_ptau = new_ptau[mask_virtual]
            part.chi[mask_virtual] = m0 / virtual_mass * part.charge_ratio[mask_virtual]
            new_ptau[mask_virtual] = (1/beta0 + old_ptau)*old_mass/virtual_mass - 1/beta0
            part.update_ptau(new_ptau)
            part.state[mask_virtual] = -VIRTUAL_ENERGY
        # Now update the parent energies
        part.add_to_energy(-E_children)

        # Add new particles
        new_part._init_random_number_generator()
        # TODO: we kill massless or neutral particles as Xsuite is not ready to handle them.
        new_part.state[massless | neutral] = -MASSLESS_OR_NEUTRAL

        # Set the state of excited ions - not supported (will fail in BDSIM when resent)
        mask = (new_part.pdg_id > 999999999) & ((new_part.pdg_id % 10) != 0)
        new_part.state[mask] = -EXCITED_ION_STATE

        part.add_particles(new_part)

    # Kill all flagged particles
    part.state[part.state==-LOST_ON_GEANT4_COLL] = LOST_ON_GEANT4_COLL
    part.state[part.state==-VIRTUAL_ENERGY] = VIRTUAL_ENERGY
    part.state[part.state==-EXCITED_ION_STATE] = EXCITED_ION_STATE
    part.state[part.state==-MASSLESS_OR_NEUTRAL] = MASSLESS_OR_NEUTRAL

    # Ensure no leftover states
    assert np.sum(part.state==LOST_WITHOUT_SPEC) == 0

    # Reshuffle
    part.reorganize()

    # Give all particles the same s position to avoid numerical differences
    part.s[part.state > 0] = s_in + coll.length



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from contextlib import contextmanager

import xobjects as xo
import xtrack as xt

from .base import BaseCollimator, BaseCrystal
from ..general import _pkg_root
from ..scattering_routines.geant4 import Geant4Engine, track_pre, track_core, track_post
from ..materials import _DEFAULT_MATERIAL, _resolve_material


class Geant4Collimator(BaseCollimator):
    _xofields = BaseCollimator._xofields | {
        'geant4_id': xo.String,
        '_tracking': xo.Int8,
        '_acc_ionisation_loss':  xo.Float64,  # TODO: this is not very robust, for when a track is done with new particles etc
    }

    isthick = True
    allow_track = True
    iscollective = True
    behaves_like_drift = True
    allow_rot_and_shift = False
    allow_loss_refinement = True
    skip_in_loss_location_refinement = True

    _depends_on = [BaseCollimator, Geant4Engine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','geant4_collimator.h')
    ]

    _noexpr_fields         = {*BaseCollimator._noexpr_fields, 'material'}
    _skip_in_to_dict       = [*BaseCollimator._skip_in_to_dict, '_tracking', '_acc_ionisation_loss']
    _store_in_to_dict      = [*BaseCollimator._store_in_to_dict, 'material']
    _internal_record_class = BaseCollimator._internal_record_class

    _allowed_fields_when_frozen = ['_tracking', '_acc_ionisation_loss']

    def __new__(cls, *args, **kwargs):
        with cls._in_constructor():
            self = super().__new__(cls, *args, **kwargs)
        return self

    def __init__(self, **kwargs):
        import xcoll as xc
        if xc.geant4.engine.is_running():
            raise ValueError('Cannot create Geant4Collimator while engine is running.')
        with self.__class__._in_constructor(self):
            to_assign = {}
            if '_xobject' not in kwargs:
                kwargs.setdefault('_tracking', True)
                kwargs.setdefault('_acc_ionisation_loss', -1.)
                kwargs.setdefault('geant4_id', ''.ljust(16))
                to_assign['name'] = xc.geant4.engine._get_new_element_name()
                to_assign['material'] = kwargs.pop('material', None)
                kwargs['_material'] = _DEFAULT_MATERIAL
            super().__init__(**kwargs)
            for key, val in to_assign.items():
                setattr(self, key, val)
            if not hasattr(self, '_equivalent_drift'):
                self._equivalent_drift = xt.Drift(length=self.length)
                self._equivalent_drift.model = 'exact'

    @property
    def angle(self):
        return BaseCollimator.angle.fget(self)

    @angle.setter
    def angle(self, val):
        if hasattr(val, '__iter__'):
            raise ValueError('The Geant4 scattering engine does not '
                           + 'support unequal jaw rotation angles')
        BaseCollimator.angle.fset(self, val)


    @property
    def material(self):
        if self._material != _DEFAULT_MATERIAL:
            return self._material

    @material.setter
    def material(self, material):
        material = _resolve_material(material, ref='geant4')
        if self.material != material:
            self._material = material


    def enable_scattering(self):
        import xcoll as xc
        xc.geant4.environment.assert_environment_ready()
        if not xc.geant4.engine.is_running():
            raise RuntimeError("Geant4 engine is not running.")
        super().enable_scattering()

    def track(self, part):
        if track_pre(self, part):
            # super().track(part)
            track_core(self, part)
            track_post(self, part)
        else:
            self._drift(part)

    def _drift(self, particles, length=None):
        if length is None:
            length = self.length
        if length != self.length:
            old_length = self._equivalent_drift.length
            self._equivalent_drift.length = length
        self._equivalent_drift.track(particles)
        if length != self.length:
            self._equivalent_drift.length = old_length

    def __setattr__(self, name, value):
        import xcoll as xc
        if name not in self._allowed_fields_when_frozen \
        and xc.geant4.engine.is_running():
            raise ValueError('Engine is running; Geant4Collimator is frozen.')
        super().__setattr__(name, value)

    @classmethod
    @contextmanager
    def _in_constructor(cls, self=None):
        original_setattr = cls.__setattr__
        if self is not None:
            self._being_constructed_ = True
        def new_setattr(self, *args, **kwargs):
            return super().__setattr__( *args, **kwargs)
        cls.__setattr__ = new_setattr
        try:
            yield
        finally:
            cls.__setattr__ = original_setattr
            if self is not None:
                self._being_constructed_ = False

    def _being_constructed(self):
        if hasattr(self, '_being_constructed_'):
            return self._being_constructed_
        else:
            return False


class Geant4CollimatorTip(Geant4Collimator):

    _xofields = Geant4Collimator._xofields | {
        'tip_thickness': xo.Float64
    }

    isthick = True
    allow_track = True
    iscollective = True
    behaves_like_drift = True
    skip_in_loss_location_refinement = True

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements', 'elements_src', 'geant4_collimator_tip.h')
    ]

    _depends_on = [*Geant4Collimator._depends_on]

    _noexpr_fields         = {*Geant4Collimator._noexpr_fields, 'tip_material'}
    _skip_in_to_dict       = [*Geant4Collimator._skip_in_to_dict, '_tracking', '_acc_ionisation_loss']
    _store_in_to_dict      = [*Geant4Collimator._store_in_to_dict, 'tip_material']
    _internal_record_class = Geant4Collimator._internal_record_class

    def __new__(cls, *args, **kwargs):
        with cls._in_constructor():
            self = super().__new__(cls, *args, **kwargs)
        return self

    def __init__(self, **kwargs):
        import xcoll as xc
        if xc.geant4.engine.is_running():
            raise ValueError('Cannot create Geant4CollimatorTip while engine is running.')
        with self.__class__._in_constructor():
            to_assign = {}
            if '_xobject' not in kwargs:
                to_assign['tip_material'] = kwargs.pop('tip_material', None)
                kwargs['_tip_material'] = _DEFAULT_MATERIAL
            super().__init__(**kwargs)
            for key, val in to_assign.items():
                setattr(self, key, val)

    @property
    def tip_material(self):
        if self._tip_material != _DEFAULT_MATERIAL:
            return self._tip_material

    @tip_material.setter
    def tip_material(self, tip_material):
        tip_material = _resolve_material(tip_material, ref='geant4')
        if self.tip_material != tip_material:
            self._tip_material = tip_material


class Geant4Crystal(BaseCrystal):
    def __init__(self, **kwargs):
        import xcoll as xc
        if xc.geant4.engine.is_running():
            raise ValueError('Cannot create Geant4Crystal while engine is running.')
        raise NotImplementedError("Geant4Crystal not yet implemented.")
